package com.sillycycle.bagleyd.life4d;

/*
 * @(#)Life4DBoxes.java
 *
 * Copyright 2010 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import javax.swing.JComboBox;
import javax.swing.JPanel;

import com.sillycycle.bagleyd.util.CheckBoxes;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComboBox;

/**
 * The <code>Life4DBoxes</code> class holds the Life configuration
 * for the <code>Life4DDialog</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Life4DBoxes extends JPanel {
	private static final long serialVersionUID = 42L;

	static final int NUM_BOXES = 20;
	static int neighborIndexValue = Life4DInterface.DEFAULT_NEIGHBOR_INDEX;
	/* default set up so ruleIndexValue will not be null */
	static int ruleIndexValue = Life4DRules.defaultRuleIndex(neighborIndexValue);
	static String ruleName = Life4DRules.parseRuleFull(Life4DRules.getRuleName(neighborIndexValue, ruleIndexValue));
	static int birthValue = Life4DRules.getBirth(ruleName);
	static int survivalValue = Life4DRules.getSurvival(ruleName);
	transient Life4DRules lifeRules = new Life4DRules(Life4DRules.indexToNeighbor[neighborIndexValue],
		birthValue, survivalValue);

	ComboBox neighborComboBox = new ComboBox("Neighborhood Modes",
		Life4DRules.neighborIndexStrings);
	CheckBoxes birthBoxes = new CheckBoxes("Birth Rule",
		Life4DRules.indexToNeighbor[neighborIndexValue],
		birthValue, NUM_BOXES);
	CheckBoxes survivalBoxes = new CheckBoxes("Survival Rule",
		Life4DRules.indexToNeighbor[neighborIndexValue],
		survivalValue, NUM_BOXES);
	ComboBox ruleComboBox = new ComboBox("Rule",
		Life4DRules.getStrings(neighborIndexValue));

	public Life4DBoxes() {
		JPanel panel = new JPanel();

		panel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		neighborComboBox.setSelectedItem(neighborIndexValue);
		panel.add(neighborComboBox);
		ruleComboBox.setSelectedItem(ruleIndexValue);
		panel.add(birthBoxes, BorderLayout.PAGE_START);
		panel.add(survivalBoxes, BorderLayout.PAGE_START);
		panel.add(ruleComboBox);
		add(BorderLayout.PAGE_START, panel);
		//pack();
		neighborComboBox.addActionListener(
			new NeighborComboBoxActionListener());
		ruleComboBox.addActionListener(
			new RuleComboBoxActionListener());
		birthBoxes.addItemListener(new BirthListener());
		survivalBoxes.addItemListener(new SurvivalListener());
	}

	class NeighborComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int oldValue = neighborIndexValue;
			//int neighbor;
			//String neighborString = (String) ((JComboBox)event.
			//	getSource()).getSelectedItem();

			neighborIndexValue = neighborComboBox.getSelectedItem();
			//for (neighbor = 0;
			//		neighbor < Life4DInterface.NEIGHBOR_GROUPS;
			//		neighbor++) {
			//	if (neighborString.indexOf(Life4DRules.neighborIndexStrings[neighbor]) >= 0) {
			//		neighborIndexValue = neighbor;
			//		break;
			//	}
			//}
			if (neighborIndexValue != -1 && oldValue != neighborIndexValue) {
				//frame.setNeighborIndex(neighborIndexValue);
				//frame.requestFocus();
				birthBoxes.updateBounds(
					Life4DRules.indexToNeighbor[neighborIndexValue]);
				survivalBoxes.updateBounds(
					Life4DRules.indexToNeighbor[neighborIndexValue]);

				Integer ruleIndex = Life4DRules.defaultRuleIndex(neighborIndexValue);
				String calculatedRuleName;

				if (ruleIndex == null) {
					ruleIndexValue = 0;
					calculatedRuleName = Life4DRules.parseRuleFull(Life4DRules.defaultRuleStrings[neighborIndexValue]);
				} else {
					ruleIndexValue = ruleIndex;
					//((JComboBox)event.getSource()).setSelectedIndex(ruleIndex);
					calculatedRuleName = Life4DRules.parseRuleFull(Life4DRules.getRuleName(neighborIndexValue, ruleIndexValue));
				}
				setRuleIndex(ruleIndexValue);
				birthValue = Life4DRules.getBirth(calculatedRuleName);
				survivalValue = Life4DRules.getSurvival(calculatedRuleName);
				birthBoxes.setValue(birthValue);
				survivalBoxes.setValue(survivalValue);
				ruleComboBox.updateList(Life4DRules.getStrings(neighborIndexValue),
					ruleIndexValue);
			}
		}
	}

	class RuleComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			String[] ruleString = null;

			String menuSelection = ((String) ((JComboBox<Object>)event.
				getSource()).getSelectedItem());

			if (menuSelection != null) {
				ruleString = menuSelection.split(",");
				// non hash way
				/*for (int ruleIndexValue = 0;
						ruleIndexValue < Life4DRules.lifeRules[neighborIndexValue].length;
						ruleIndexValue++) {
					if (ruleString[0].equals(Life4DRules.lifeRules[neighborIndexValue][ruleIndexValue].getRuleName())) {*/
				Integer rule = lifeRules.getRuleIndex(neighborIndexValue, ruleString[0]);
				if (rule != null) {
					int ruleIndex = rule;

					birthBoxes.setValue(Life4DRules.lifeRules[neighborIndexValue][ruleIndex].getBirth());
					survivalBoxes.setValue(Life4DRules.lifeRules[neighborIndexValue][ruleIndex].getSurvival());
				}
			}
			if (ruleString == null/* || "Other".equals(ruleString[0])*/) {
				Integer ruleIndex = Life4DRules.defaultRuleIndex(neighborIndexValue);

				if (ruleIndex == null) {
					ruleName = Life4DRules.parseRuleFull(Life4DRules.defaultRuleStrings[neighborIndexValue]);
				} else {
					ruleIndexValue = ruleIndex;
					//((JComboBox)event.getSource()).setSelectedIndex(ruleIndex);
					ruleName = Life4DRules.parseRuleFull(Life4DRules.getRuleName(neighborIndexValue, ruleIndexValue));
				}
				birthValue = Life4DRules.getBirth(ruleName);
				survivalValue = Life4DRules.getSurvival(ruleName);
			}
		}
	}

	class BirthListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			birthValue = birthBoxes.getValue();
			setMenu(neighborIndexValue,
				birthValue, survivalValue);
			/*frame.setRules();
			frame.requestFocus();*/
		}
	}

	class SurvivalListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			survivalValue = survivalBoxes.getValue();
			setMenu(neighborIndexValue,
				birthValue, survivalValue);
			/*frame.setRules();
			frame.requestFocus();*/
		}
	}

	public void addActionListener(ActionListener l) {
		neighborComboBox.addActionListener(l);
		neighborIndexValue = neighborComboBox.getSelectedItem();
		Integer ruleIndex = Life4DRules.defaultRuleIndex(neighborIndexValue);

		if (ruleIndex == null) {
			ruleName =  Life4DRules.parseRuleFull(Life4DRules.defaultRuleStrings[neighborIndexValue]);
		} else {
			ruleIndexValue = ruleIndex;
			ruleName = Life4DRules.parseRuleFull(Life4DRules.getRuleName(neighborIndexValue, ruleIndexValue));
		}
		birthValue = Life4DRules.getBirth(ruleName);
		survivalValue = Life4DRules.getSurvival(ruleName);
	}

	public void addItemListener(ItemListener l) {
		birthBoxes.addItemListener(l);
		survivalBoxes.addItemListener(l);
		birthValue = birthBoxes.getValue();
		survivalValue = survivalBoxes.getValue();
	}

	public void setNeighborIndex(int value) {
		Life4DDialog.neighborIndexValue = value;
		neighborComboBox.setSelectedItem(value);
	}

	public void setRuleIndex(int value) {
		ruleComboBox.setSelectedItem(value);
	}

	public void setBirth(int value) {
		birthBoxes.setValue(value);
	}

	public void setSurvival(int value) {
		survivalBoxes.setValue(value);
	}

	public void updateList() {
		if (Life4DRules.defaultRuleIndex(neighborIndexValue) != null)
			ruleComboBox.updateList(Life4DRules.getStrings(neighborIndexValue),
				Life4DRules.defaultRuleIndex(neighborIndexValue));
	}

	public void setMenu(int neighborIndexValue,
			int birthValue, int survivalValue) {
		String ruleString = Life4DRules.parseRuleCondense("B" + birthBoxes.ruleBuilder(birthValue) +
			"/S" + survivalBoxes.ruleBuilder(survivalValue));
//System.out.println("1LifeBoxes setMenu: " + neighborIndexValue + ", " + ruleString);
		Integer ri = lifeRules.getRuleIndex(neighborIndexValue, ruleString);

//System.out.println("2LifeBoxes setMenu: " + ri);
		//for (i = 0; i < MAX_GROUP - FIRST_GROUP; i++)
		if (ri == null) {
			ruleComboBox.setSelectedItem(
				Life4DRules.lifeRules[neighborIndexValue].length);
		} else {
//System.out.println("3LifeBoxes setMenu " + ri.intValue());
			ruleComboBox.setSelectedItem(ri);
		}
	}

	public int getNeighborIndex() {
		return neighborComboBox.getSelectedItem();
	}

	public int getRuleIndex() {
		return ruleComboBox.getSelectedItem();
	}

	public int getBirth() {
		return birthBoxes.getValue();
	}

	public int getSurvival() {
		return survivalBoxes.getValue();
	}
}
