package com.sillycycle.bagleyd.life4d;

/*
 * @(#)Life4DCanvas.java
 *
 * Copyright 2010 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.util.Box4D;
import com.sillycycle.bagleyd.util.Cell4D;
import com.sillycycle.bagleyd.util.CellInfo;
import com.sillycycle.bagleyd.util.Dimension4D;
import com.sillycycle.bagleyd.util.Graphics3D;
import com.sillycycle.bagleyd.util.Point4D;
import com.sillycycle.bagleyd.util.Signatures;

/**
 * The <code>Life4DCanvas</code> class does all the graphics for
 * <code>Life4DFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Life4DCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");

	static final int MAX_STATES = 4;
	static final int MAX_AGE = 127;
	static final String DATAFILE = "data.life";

	Random generator;
	int startSpeed = Life4DInterface.DEFAULT_SPEED;
	int speed = Life4DInterface.DEFAULT_SPEED;
	int columns = Life4DInterface.DEFAULT_COLUMNS;
	int rows = Life4DInterface.DEFAULT_ROWS;
	int stacks = Life4DInterface.DEFAULT_STACKS;
	int zones = Life4DInterface.DEFAULT_ZONES;
	int ncols = columns;
	int nrows = rows;
	int nstacks = stacks;
	int nzones = zones;
	int minCol = columns / 2 - 1;
	int maxCol = columns / 2 + 1;
	int minRow = rows / 2 - 1;
	int maxRow = rows / 2 + 1;
	int minStack = stacks / 2 - 1;
	int maxStack = stacks / 2 + 1;
	int minZone = zones / 2 - 1;
	int maxZone = zones / 2 + 1;
	int area = ncols * nrows;
	int volume = area * nstacks;
	int mouseDepth = nstacks / 2;
	int mouseZone = nzones / 2;
	boolean randomOrient3D = true;
	boolean randomOrient = false;
	boolean grid = false;
	boolean lissajous = true;
	boolean wrap = true;
	boolean visible = true;
	boolean mono = false;
	boolean round = false;
	boolean wireframe = false;
	boolean findStable = false;
	boolean unboundDetection = false;
	boolean edgeDetection = false;
	//boolean focus = true;
	boolean mouseDown = false;
	boolean edge = false;
	boolean paused = true;
	boolean generatingDB = true;
	int neighbors =
		Life4DRules.indexToNeighbor[Life4DInterface.DEFAULT_NEIGHBOR_INDEX];
	int birth = 0;
	int survival = 0;
	int excludeCount = Life4DInterface.DEFAULT_EXCLUDE_COUNT;
	int searchCount = Life4DInterface.DEFAULT_SEARCH_COUNT;
	int excludePeriod = Life4DInterface.DEFAULT_EXCLUDE_PERIOD;
	int searchLength = Life4DInterface.DEFAULT_SEARCH_LENGTH;
	int fillPercent = Life4DInterface.DEFAULT_FILL_PERCENT;
	int blobPercent = Life4DInterface.DEFAULT_BLOB_PERCENT;
	int stirs = Life4DInterface.DEFAULT_STIRS;
	Color initialCellColor = Color.red;
	Color gridColor = Color.gray;
	Color background = Color.black;
	Cell4D cellPl;
	transient Graphics3D g4d;
	transient Capturer capturer;
	JFrame frame = null;
	boolean debug = false;
	String name = "newbie";
	Life4DMatrix field = null;
	transient Life4DSoup soup = null;
	int gen = 0, genOffset = 0;
	int nCells = 0, oldnCells = 0, cCells = 0;
	int states = 2;
	String stringSave = null;
	transient Life4DRules param = null;
	String signature = null;
	String stamp = null;
	transient Signatures signatures = null;
	transient Signatures findSignatures = null;
	int[] findNCells = null;
	int[] findPeriod = null;
	int[] findExcludePeriod = null;
	int checkFind = -1;
	StringBuilder initSB = new StringBuilder();

	private synchronized void pulse(boolean draw, boolean stirring) {
		cCells = 0;
		eraseFrame();
		/* 1. */
		/* for each live cell in linked list
		 *   create a dead list if toggle 0 and field array 0
		 * set toggle in field */
		if (param.getBirth(0)) {
			for (int position = 0; position < getColumns() * getRows(); position++) {
				if (field.getCell(position) == 0) {
					field.setDeadCell(position);
				}
			}
		} else {
			// causing ConcurrentModificationException
			//for (Iterator<CellInfo> iterator = field.list.iterator();
			//		iterator.hasNext();) {
			//	CellInfo cellInfo = iterator.next();
			//}
			for (int i = 0; i < field.list.size(); i++) {
				CellInfo cellInfo = field.list.get(i);
				int position = cellInfo.getPosition();
				int col = getColumn(position);
				int row = getRow(position);
				int stack = getStack(position);
				int zone = getZone(position);
				for (int neighbor = 0; neighbor < Math.abs(neighbors); neighbor++)
					field.setDeadCell(positionOfNeighbor(neighbor, col, row, stack, zone));
			}
		}
		/* 2. */
		/* for each dead cell: reset toggles back to 0;
		 *   if comes alive keep, else remove */
		for (Iterator<CellInfo> iterator = field.deadList.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int deadPosition = cellInfo.getPosition();
			int col = getColumn(deadPosition);
			int row = getRow(deadPosition);
			int stack = getStack(deadPosition);
			int zone = getZone(deadPosition);
			int fertility = getNeighbors(col, row, stack, zone);
			byte state;

			if (stirring && stirs > gen) {
				state = (byte) ((param.getBirthStir(fertility)) ?
					(states - 1) : 0);
			} else {
				state = (byte) ((param.getBirth(fertility)) ?
					(states - 1) : 0);
			}
			field.resetCellBit(deadPosition);
			if (state == 0) {
				iterator.remove();
			}
		}
		/* 3. */
		/* for each live cell set toggle if dies */
		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int col = getColumn(position);
			int row = getRow(position);
			int stack = getStack(position);
			int zone = getZone(position);
			int value = getNeighbors(col, row, stack, zone);
			byte nextState;
			int extra = 0;

			if (states > 2) {
				byte state;
				extra = 1;
				state = cellInfo.getCell();
				if (state == 3) {
					if (stirring && stirs > gen)
						nextState = (byte) ((param.getSurvivalStir(value +
							extra)) ? (states - 1) : 0);
					else
						nextState = (byte) ((param.getSurvival(value +
							extra)) ? (states - 1) : 0);
					if (nextState != state)
						cellInfo.setTemp((byte) (nextState + 1)); // come back later and set cell to 0
				} else { //if (state == 1 or 2)
					byte nextState0, nextState1;

					if (stirring && stirs > gen)
						nextState0 = (byte) ((param.getBirthStir(value + 1)) ? 1 : 0);
					else
						nextState0 = (byte) ((param.getBirth(value + 1)) ? 1 : 0);
					if (stirring && stirs > gen)
						nextState1 = (byte) ((param.getSurvivalStir(value)) ? 1 : 0);
					else
						nextState1 = (byte) ((param.getSurvival(value)) ? 1 : 0);
					if (state == 1)
						nextState = (byte) (((nextState1 << 1) & 0xff) + nextState0);
					else
						nextState = (byte) (((nextState0 << 1) & 0xff) + nextState1);
					if (nextState != state)
						cellInfo.setTemp((byte) (nextState + 1)); // come back later and set cell to 0
				}
			} else {
				if (stirring && stirs > gen)
					nextState = (byte) ((param.getSurvivalStir(value +
						extra)) ? (states - 1) : 0);
				else
					nextState = (byte) ((param.getSurvival(value +
						extra)) ? (states - 1) : 0);
				if (nextState == 0)
					cellInfo.setTemp((byte) 1); // come back later and set cell to 0
			}
		}
		/* 4. */
		/* for each live cell update according to toggle
		 *   if dies erase cell remove.
		 *   if lives update age, draw if necessary. */
		for (Iterator<CellInfo> iterator = field.list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int position = cellInfo.getPosition();
			byte temp = cellInfo.getTemp();
			switch (temp) {
			case 0:
			    if (states == 2) {
				byte age = cellInfo.getCell();
				if (!mono && draw && age != MAX_AGE) {
				    age = (byte) (age + 1);
				    cellInfo.setCell(age);
				    /*drawCell(cellInfo);*/
				    field.setArrayCell(position, age);
				}
				//cellInfo.setCell(field.getCell(position));
			    }
			    break;
			case 1:
			    /*if (draw) {
			    eraseCell(cellInfo);
			    }*/
			    field.setArrayCell(position, (byte) 0);
			    nCells--;
			    cCells++;
			    iterator.remove();
			    break;
			default:
			    temp = (byte) (temp - 1);
			    cellInfo.setTemp((byte) 0);
			    cellInfo.setCell(temp);
			    field.setArrayCell(position, temp);
			    /*if (draw) {
			    drawCell(cellInfo);
			    }*/
			    cCells++;
			    break;
			}
		}
		/* 5. */
		/* for each dead cell update to live cell. */
		for (Iterator<CellInfo> iterator = field.deadList.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int deadPosition = cellInfo.getPosition();
			int col = getColumn(deadPosition);
			int row = getRow(deadPosition);
			int stack = getStack(deadPosition);
			int zone = getZone(deadPosition);
			byte nextState = (byte) (states - 1);
			/*if (draw)
				drawCell(cellInfo);*/
			field.setCell(deadPosition, nextState);
			checkAliveBounds(col, row, stack, zone);
			nCells++;
			cCells++;
			iterator.remove();
		}
		if (nCells > 0) {
			gen++;
			if (gen - stirs == 1 && unboundDetection)
				saveInitLife();
		} else {
			paused = true;
		}
		/* 6. */
		if (draw)
			showLiveCells();
	}

	void lifePulse(boolean draw, boolean fast, boolean stirring) {
		if (lissajous)
			g4d.lissajous();
		pulse(draw, stirring);
		if (draw && (!fast || (gen & 63) == 0)) {
			((Life4DFrame) frame).updateGen(gen);
			((Life4DFrame) frame).updateCells(nCells);
		}
	}

	public Life4DCanvas() {
		generator = new Random(System.nanoTime());
		capturer = new Capturer();
		soup = new Life4DSoup(this);
	}

	void inits() {
		resizeField();
	}

	void ruleChange() {
		if ((Life4DFrame) frame != null) {
			param = new Life4DRules(neighbors,
				((Life4DFrame) frame).getBirth(),
				((Life4DFrame) frame).getSurvival());
			String ruleName = param.getRuleName();

			if (ruleName == null)
				((Life4DFrame) frame).updateRule(Integer.toString(neighbors));
			else
				((Life4DFrame) frame).updateRule(neighbors + ruleName);
			if (debug)
				System.out.println("ruleChange: " + ruleName);
		}
		signatures = null;
	}

	public int getStates() {
		return this.states;
	}

	public int getNeighborIndex() {
		return Life4DRules.neighborToIndex(neighbors);
	}

	Integer getRuleIndex(int neighborIndex) {
		String ruleName = param.getRuleName();

		if (ruleName == null || ruleName.isEmpty())
			return Life4DRules.defaultRuleIndex(neighborIndex);
		ruleName = Life4DRules.parseRuleCondense(param.getRuleName());
		if (ruleName == null) {
			return null;
		}
		if (debug)
			System.out.println("getRuleIndex: " + ruleName);
		return param.getRuleIndex(neighborIndex, ruleName);
	}

	static boolean polychoronChange(int oldNeighborIndex, int newNeighborIndex) {
		if (oldNeighborIndex != newNeighborIndex) {
			int oldPolychoron = Life4DRules.neighborIndexToPolychoron[oldNeighborIndex];
			int newPolychoron = Life4DRules.neighborIndexToPolychoron[newNeighborIndex];

			return (oldPolychoron != newPolychoron);
		}
		return false;
	}

	int getPolychoron() {
		return Life4DRules.neighborToPolychoron(neighbors);
	}

	void setNeighbors(int neighborIndex) {
		neighbors = Life4DRules.indexToNeighbor[neighborIndex];
	}

	static int getRealColumns(int cols) {
		return cols;
	}

	int getStacksFromColumns(int cols) {
		int polychoron = getPolychoron();
		int calculatedStacks = cols / 2;

		if (!boundsCheck24(polychoron, cols, cols, calculatedStacks, this.nzones)) {
			calculatedStacks++;
		}
		return calculatedStacks;
	}

	int getZonesFromColumnsStacks(int cols, int stks) {
		int polychoron = getPolychoron();
		int calculatedZones = 3 * cols / 8;

		if (!boundsCheck24(polychoron, cols, cols, stks, calculatedZones)) {
			calculatedZones++;
		}
		return calculatedZones;
	}

	public int getColumns() {
		return this.ncols;
	}

	public int getRows() {
		return this.nrows;
	}

	public int getStacks() {
		return this.nstacks;
	}

	public int getZones() {
		return this.nzones;
	}

	/* Checks to make sure there is a cell there */
	static boolean cellCheck(int polychoron, int x, int y, int z, int w) {
		return (polychoron != 24 || ((x + y + z + w) & 1) == 0);
	}

	/* Checks to make sure it wraps well */
	static boolean boundsCheck24(int polychoron, int x, int y, int z, int w) {
		return (polychoron != 24 || ((x + y + z + w) & 1) == 0);
	}

	public void speedUp() {
		speed += 10;
		if (speed > Life4DInterface.MAX_SPEED)
			speed = Life4DInterface.MAX_SPEED;
		((Life4DFrame) frame).setDelay(Life4DInterface.MAX_SPEED +
			Life4DInterface.MIN_SPEED - 1 - speed);
	}

	public void slowDown() {
		speed -= 10;
		if (speed < Life4DInterface.MIN_SPEED)
			speed = Life4DInterface.MIN_SPEED;
		((Life4DFrame) frame).setDelay(Life4DInterface.MAX_SPEED +
			Life4DInterface.MIN_SPEED - 1 - speed);
	}

	void resizeField() {
		int newWidth = getSize().width;
		int newHeight = getSize().height;
		int tempCols;
		int oldArea = area;
		int oldVolume = volume;
		int polychoron = getPolychoron();

		capturer.clearCapture();
		tempCols = Life4DFrame.getColumns();
		ncols = getRealColumns(tempCols);
		nrows = ncols;
		area = ncols * nrows;
		nstacks = getStacksFromColumns(tempCols);
		mouseDepth = nstacks / 2;
		volume = area * nstacks;
		if (newWidth == 0)
			newWidth = Life4DInterface.DEFAULT_CELLSIZE *
				ncols + 4;
		if (newHeight == 0)
			newHeight = Life4DInterface.DEFAULT_CELLSIZE *
				nrows + 4;

		cellPl = new Cell4D(polychoron, ncols, nrows/*, nstacks, nzones*/,
			newWidth, newHeight/*, mouseZone*/);
//System.out.println("Resize " + neighbors);
//System.out.println("ncols " + ncols + ", " + nrows + ", " + nstacks + ", " + nzones);
//System.out.println("cols " + cols + ", " + rows + ", " + stacks + ", " + zones);
//System.out.println(param.getCommonName());
		if (field == null) {
			columns = ncols;
			rows = nrows;
			stacks = nstacks;
			zones = nzones;
			field = new Life4DMatrix(ncols, nrows, nstacks, nzones);
			gen = 0;
			genOffset = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			minStack = nstacks / 2 - 1;
			maxStack = nstacks / 2 + 1;
			minZone = nzones / 2 - 1;
			maxZone = nzones / 2 + 1;
			nCells = 0;
			if ((Life4DFrame) frame != null) {
				((Life4DFrame) frame).initPopup();
			}
		} else if (columns != ncols || rows != nrows ||
				stacks != nstacks || zones != nzones) {
			if (!boundsCheck24(polychoron, minCol, minRow, minStack, minZone)) {
				if (minStack > 0)
					minStack--;
				else if (minRow > 0)
					minRow--;
				else if (minCol > 0)
					minCol--;
			}
			recalibrateBounds(columns, rows, stacks, zones);

			int fieldX = maxCol - minCol + 1;
			int fieldY = maxRow - minRow + 1;
			int fieldZ = maxStack - minStack + 1;
			int fieldW = maxZone - minZone + 1;
			int fieldArea = fieldX * fieldY;
			int fieldVolume = fieldArea * fieldZ;
			Life4DMatrix cage = new Life4DMatrix(fieldX, fieldY, fieldZ, fieldW);

			for (CellInfo cellInfo : field.list) {
				int oldP = cellInfo.getPosition();
				int i = getColumn(oldP, columns);
				int j = getRow(oldP, columns, oldArea);
				int k = getStack(oldP, oldArea, oldVolume);
				int l = getZone(oldP, oldVolume);
				byte value = field.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i - minCol, j - minRow,
						k - minStack, l - minZone,
						fieldX, fieldArea, fieldVolume);
					cage.setArrayCell(p, value);
				}
			}
			columns = ncols;
			rows = nrows;
			stacks = nstacks;
			zones = nzones;
			field = new Life4DMatrix(ncols, nrows, nstacks, nzones);
			gen = 0;
			genOffset = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			minStack = nstacks / 2 - 1;
			maxStack = nstacks / 2 + 1;
			minZone = nzones / 2 - 1;
			maxZone = nzones / 2 + 1;
			nCells = 0;

			int offsetX = minCol + 1 - fieldX / 2;
			int offsetY = minRow + 1 - fieldY / 2;
			int offsetZ = minStack + 1 - fieldZ / 2;
			int offsetW = minZone + 1 - fieldW / 2;

			if (!boundsCheck24(polychoron, offsetX, offsetY, offsetZ, offsetW)) {
				offsetZ--;
			}
			for (int l = 0; l < fieldW; l++)
			if ((l + offsetW >= 0) && (l + offsetW <= nzones - 1))
			for (int k = 0; k < fieldZ; k++)
			if ((k + offsetZ >= 0) && (k + offsetZ <= nstacks - 1))
			for (int j = 0; j < fieldY; j++)
			if ((j + offsetY >= 0) && (j + offsetY <= nrows - 1))
			for (int i = 0; i < fieldX; i++)
			if ((i + offsetX >= 0) && (i + offsetX <= ncols - 1)) {
				int oldP = getPosition(i, j, k, l,
					fieldX, fieldArea, fieldVolume);
				byte value = cage.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i + offsetX, j + offsetY, k + offsetZ, l + offsetW,
						ncols, area, volume);
					field.setCell(p, value);
					nCells++;
				}
			}
			if ((Life4DFrame) frame != null) {
				((Life4DFrame) frame).updateGen(gen);
				((Life4DFrame) frame).updateCells(nCells);
				((Life4DFrame) frame).initPopup();
			}
		}
		g4d = new Graphics3D(this.getGraphics(),
			ncols, nrows, nstacks,/* nzones,*/
			newWidth, newHeight, mouseDepth/*, mouseZone*/);
	}

	void xorCell(CellInfo info, byte state, int k) {
		int zone = getZone(info);
		if (zone != mouseZone)
			return;
		int col = getColumn(info);
		int row = getRow(info);
		int stack = getStack(info);
		Color color;

		if (state == 0) {
			color = background;
		} else {
			if (mono) {
				color = Color.white;
			} else if (states == 2) {
				color = colorRainbow(state, 16);
			} else {
				color = colorRainbow((state - 1) * 16, 16);
			}
		}
		if ((k & 1) == 0) // Simulate Xor
			color = new Color(255 - color.getRed(),
				255 - color.getGreen(),
				255 - color.getBlue());
		cellPl.drawCell(g4d, round, wireframe, color, gridColor,
			col, row, stack/*, zone*/);
	}

	void showLiveCells() {
		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			cellInfo.setDistance(g4d.getDistance(getColumn(position), getRow(position), getStack(position)));
			/*			System.out.println("distance: " +
			col + " " + row + " " + stack + " " + zone + " : " +
			info.getDistance());*/
		}
		/* Sort live cells and display. */
		java.util.Collections.sort(field.list, new Life4DMatrix());
		for (CellInfo cellInfo : field.list) {
			drawCell(cellInfo);
		}
	}

	// Should make some attempt at minimizing size if on a border...
	void printField(StringBuilder sb) {
		String space = " ";
		String newline = NL;
		String pound = "#";
		boolean firstPoint = true;
		int i, j, k, l, p;
		int startX = 0, startY = 0, startZ = 0, startW = 0;
		int endX = 0, endY = 0, endZ = 0, endW = 0;
		int polychoron = getPolychoron();

		sb.append(pound).append("Dim").append(space)
			.append(4).append(newline);
		sb.append(pound).append("Neighbors").append(space)
			.append(neighbors).append(newline);
		sb.append(pound).append("Rule").append(space)
			.append("B").append(Life4DRules.condense(Life4DRules.getBirthString(param.getRuleName())))
			.append("S").append(Life4DRules.condense(Life4DRules.getSurvivalString(param.getRuleName()))).append(newline);
		sb.append(pound).append("Cells").append(space)
			.append(nCells).append(newline);
		sb.append(pound).append("Name").append(space)
			.append(name).append(newline);
		/*sb.append(pound).append("Sig").append(space)
			.append("An A0 <> D1 Dn").append(newline);*/
		for (l = 0; l < nzones; l++)
		for (k = 0; k < nstacks; k++)
		for (j = 0; j < nrows; j++)
		for (i = 0; i < ncols; i++) {
				p = getPosition(i, j, k, l);
				if (field.getCell(p) != 0) {
					if (firstPoint) {
						startX = i;
						startY = j;
						startZ = k;
						startW = l;
						endX = i;
						endY = j;
						endZ = k;
						endW = l;
						firstPoint = false;
					} else {
						if (i < startX)
							startX = i;
						if (j < startY)
							startY = j;
						if (k < startZ)
							startZ = k;
						if (i > endX)
							endX = i;
						if (j > endY)
							endY = j;
						if (k > endZ)
							endZ = k;
						if (l > endW)
							endW = l;
					}
				}
		}
		if (!boundsCheck24(polychoron, startX, startY, startZ, startW)) {
			if (startZ >= startX && startZ >= startY)
				startZ--;
			else if (startY >= startZ && startY >= startX)
				startY--;
			else
				startX--;
		}
		sb.append(pound + "Width" + space + (endX - startX + 1) + newline);
		sb.append(pound + "Height" + space + (endY - startY + 1) + newline);
		sb.append(pound + "Depth" + space + (endZ - startZ + 1) + newline);
		sb.append(pound + "Zone" + space + (endW - startW + 1) + newline);
		sb.append(pound + "P" + space + (startX - ncols / 2) + space +
			(startY - nrows / 2) + space +
			(startZ - nstacks / 2) + space +
			(startW - nzones / 2) + newline);
		//sb.append(pound + "P" + space + 0 + space + 0 + newline);
		for (l = startW; l <= endW; l++) {
		for (k = startZ; k <= endZ; k++) {
		for (j = startY; j <= endY; j++) {
		for (i = startX; i <= endX; i++) {
				p = getPosition(i, j, k, l);
				if (((field.getCell(p)) == 0) ||
						(!cellCheck(polychoron, i, j, k, l))) {
					sb.append(Life4DInterface.DEFAULT_DEAD_CHAR);
				} else if (states == 2) {
					sb.append(Life4DInterface.DEFAULT_ALIVE_CHAR);
				} else {
					sb.append(field.getCell(p));
				}
		}
		sb.append(newline);
		}
		sb.append(newline);
		}
		sb.append(newline);
		}
	}

	void clearField(boolean draw) {
		field.clearLiveCells();
		if (draw)
			eraseFrame();
		gen = 0;
		genOffset = 0;
		param.setStirs();
		edge = false;
		minCol = ncols / 2 - 1;
		maxCol = ncols / 2 + 1;
		minRow = nrows / 2 - 1;
		maxRow = nrows / 2 + 1;
		minStack = nstacks / 2 - 1;
		maxStack = nstacks / 2 + 1;
		minZone = nzones / 2 - 1;
		maxZone = nzones / 2 + 1;
		nCells = 0;
		if (draw && (Life4DFrame) frame != null) {
			//Graphics g = this.getGraphics();

			((Life4DFrame) frame).updateGen(gen);
			((Life4DFrame) frame).updateCells(nCells);
			//drawFrame(g, true);
			if (grid)
				drawGrid();
		}
	}

	void showMessage(String msg) {
		((Life4DFrame) frame).showMessage(msg);
	}

	/*void setCell(int col, int row, int stack, int zone) {
		int polychoron = getPolychoron();

		if (col >= 0 && col < ncols &&
				row >= 0 && row < nrows &&
				stack >= 0 && stack < nstacks &&
				zone >= 0 && zone < nzones) {
			setCell(col, row, stack, zone, (byte) 1);
			if (!cellCheck(polychoron, col, row, stack, zone))
				throw new IllegalArgumentException("failed cell check");
		}
	}*/

	void setCell(int col, int row, int stack, int zone, int state) {
		int polychoron = getPolychoron();

		if (col >= 0 && col < ncols &&
				row >= 0 && row < nrows &&
				stack >= 0 && stack < nstacks &&
				zone >= 0 && zone < nzones) {
			setCell(col, row, stack, zone, (byte) state);
			if (!cellCheck(polychoron, col, row, stack, zone))
				throw new IllegalArgumentException("failed cell check");
		}
	}

	public String getGridStamp() {
		return findGridStamp(new Box4D(minCol, minRow, minStack, minZone,
			maxCol - minCol + 1,
			maxRow - minRow + 1,
			maxStack - minStack + 1,
			maxZone - minZone + 1));
	}

	/** Counts the number of live cells and adds the coordinates.
	 * Not as pretty as a signature but usually catches cases
	 * where life form just is reoriented.
	 * @param b
	 * @return grid stamp string
	 */
	public String findGridStamp(Box4D b) {
		int p;
		int polychoron = getPolychoron();
		Point4D linear = new Point4D(0, 0, 0, 0);
		Point4D square = new Point4D(0, 0, 0, 0);

		for (int l = b.w; l < b.w + b.zone; l++)
		for (int k = b.z; k < b.z + b.depth; k++)
		for (int j = b.y; j < b.y + b.height; j++)
		for (int i = b.x; i < b.x + b.width; i++)
		if (cellCheck(polychoron, i, j, k, l)) {
				p = getPosition(i, j, k, l);
				if (field.getCell(p) != 0) {
					int col = i - ncols / 2;
					int row = j - nrows / 2;
					int stack = k - nstacks / 2;
					int zone = l - nzones / 2;

					linear.x += col;
					linear.y += row;
					linear.z += stack;
					linear.w += zone;
					square.x += col * col;
					square.y += row * row;
					square.z += stack * stack;
					square.w += zone * zone;
				}
		}
		return "x" + linear.x + "y" + linear.y + "z" + linear.z + "w" + linear.w +
			"X" + square.x + "Y" + square.y + "Z" + square.z + "W" + square.w;
	}

	public String getGridSignature() {
		return findGridSignature(new Box4D(minCol, minRow, minStack, minZone,
			maxCol - minCol + 1,
			maxRow - minRow + 1,
			maxStack - minStack + 1,
			maxZone - minZone + 1));
	}

	/** Counts the number of live and dead cells with a certain
	 * number of live neighbors and then passes these "bucket"
	 * counts to another function which returns a formatted
	 * signature code for all cells currently in the grid.
	 * @param b
	 * @return grid signature string
	 */
	public String findGridSignature(Box4D b) {
		int neighborCount, p;
		int polychoron = getPolychoron();
		// 80 not used in Life4DRules due to # of bits in Integer
		LifeSig sigCount = new LifeSig(80/*Life4DRules.MAX_NEIGHBORS*/);
		int extraX = 0;
		int extraY = 0;
		int extraZ = 0;
		int extraW = 0;

		if (polychoron == 16 && b.w > 0) { // need a little more buffer
			extraW = 1;
		}
		if (neighbors == -32) { // need a little more buffer too
			if (b.x > 0)
				extraX = 1;
			if (b.y > 0)
				extraY = 1;
			if (b.z > 0)
				extraZ = 1;
			if (b.w > 0)
				extraW = 1;
		}

		for (int l = b.w - extraW; l < b.w + b.zone + 2 * extraW; l++)
		for (int k = b.z - extraZ; k < b.z + b.depth + 2 * extraZ; k++)
		for (int j = b.y - extraY; j < b.y + b.height + 2 * extraY; j++)
		for (int i = b.x - extraX; i < b.x + b.width + 2 * extraX; i++)
		if (cellCheck(polychoron, i, j, k, l)) {
				int newI = (i + columns) % columns;
				int newJ = (j + rows) % rows;
				int newK = (k + stacks) % stacks;
				int newL = (l + zones) % zones;

				p = getPosition(newI, newJ, newK, newL);
				neighborCount = getNeighbors(newI, newJ, newK, newL);
				if (field.getCell(p) == 0) {
					sigCount.dead++;
					sigCount.deadCount[neighborCount]++;
				} else {
					sigCount.live++;
					sigCount.liveCount[neighborCount]++;
				}
		}
		return sigCount.signatureCode();
	}

	public boolean getSignatureList(boolean print,
			boolean db, boolean special, String species) {
		final int MAX_OSC = 72;
		final int MAX_SMOKE = 8;
		int osc = 0;
		int numCells;

		if (signatures == null) {
			signatures = new Signatures();
		}
		String[] gSignature = new String[MAX_OSC];
		gSignature[0] = getGridSignature();
		numCells = Signatures.signatureToNumberLiveCells(gSignature[0]);
		if (print && numCells != 0) {
			if (db) {
				if (species != null)
					System.out.print(species);
				System.out.print("[" + numCells + "]|");
			}
			System.out.println(gSignature[0]);
		}
		// This stops at first match.
		// This is not guaranteed to get all signatures
		// for example picture 2 clock like life forms
		// spinning opposite each other, they could have
		// same overall signature until they point at each
		// other.  Life forms can probably be entered in DB
		// in such a way that this is alleviated.

		if (special) {
			signatures.putSignature(gSignature[0]);
			return true;
		}
		while (osc < MAX_OSC - 1) {
			lifePulse(!db, true, false);
			String signatureN = getGridSignature();
			numCells = Signatures.signatureToNumberLiveCells(signatureN);
			gSignature[osc + 1] = signatureN;
			for (int smoke = 0; smoke < MAX_SMOKE; smoke++) {
				if (osc >= smoke && signatureN.equals(gSignature[smoke])) {
					for (int i = 0; i <= osc; i++) {
						signatures.putSignature(gSignature[i]);
					}
					return true;
				}
			}
			if (osc == MAX_OSC || numCells == 0 || "0 <> 0".equals(signatureN)) {
				if (species == null) // Screen junk
					return true;
				/*if (debug)*/
					System.out.println(species +
						": failed on osc " +
						osc + ", " + signatureN);
				return false;
			}
			if (print) {
				if (db) {
					if (species != null)
						System.out.print(species);
					System.out.print("[" + numCells + "]|");
				}
				System.out.println(signatureN);
			}
			osc++;
		}
		return false;
	}

	static int[] toIntArray(ArrayList<Integer> list) {
		int[] ret = new int[list.size()];
		int i = 0;
		for (Integer nonprim : list)
			ret[i++] = nonprim;
		return ret;
	}

	private void addSignatures(String string, boolean print) {
		String fileName = "sig/" + string + neighbors + "B" +
			Life4DRules.getBirthString(param.getRuleName()) + "S" +
			Life4DRules.getSurvivalString(param.getRuleName()) + ".sig";
		try (BufferedReader br = new BufferedReader(
				new FileReader(fileName))) {
			String lineData, sig;
			ArrayList<Integer> findNCellsList = null;
			ArrayList<Integer> findPeriodList = null;
			ArrayList<Integer> findExcludePeriodList = null;
			int mult;

			while ((lineData = br.readLine()) != null) {
				String[] lineToken;

				switch (string) {
				case "junk":
					sig = lineData;
					mult = signatures.containsSignature(sig);
					if (mult == 0) {
						signatures.putSignature(sig);
						if (print) {
							int numCells = Signatures.signatureToNumberLiveCells(sig);

							System.out.println(fileName + "[" + numCells + "]|" + sig);
						}
					} else {
						System.out.println(string + "(" + mult + "): " + sig);
					}
					break;
				case "find":
					lineToken = lineData.split("\\|", 2);
					if (lineToken.length > 1) {
						sig = lineToken[1];
						mult = signatures.containsSignature(sig);
						if (mult != 0)
							System.out.println(string + "(" + mult + "): " + sig);
						findSignatures.putSignature(sig);
						if (print) {
							int numCells = Signatures.signatureToNumberLiveCells(sig);
							System.out.println(lineToken[0] + "[" + numCells + "]|" + sig);
						}
					} else {
						String[] lineTokenParen = lineData.split("\\(p", 2);
						if (lineTokenParen.length > 1) {
							String[] lineTokenParen2 = lineTokenParen[1].split("\\)", 2);
							if (findPeriodList == null)
								findPeriodList = new ArrayList<>();
							findPeriodList.add(Integer.parseInt(lineTokenParen2[0].trim()));
						}
						String[] lineTokenBracket = lineData.split("\\[", 2);
						if (lineTokenBracket.length > 1) {
							String[] lineTokenBracket2 = lineTokenBracket[1].split("\\]", 2);
							if (findNCellsList == null)
								findNCellsList = new ArrayList<>();
							findNCellsList.add(Integer.parseInt(lineTokenBracket2[0].trim()));
						}
						String[] lineTokenOtherBracket = lineData.split("\\]", 2);
						if (lineTokenOtherBracket.length > 0) {
							if (findExcludePeriodList == null)
								findExcludePeriodList = new ArrayList<>();
							findExcludePeriodList.add(Integer.parseInt(lineTokenOtherBracket[1].trim()));
						}
					}
					break;
				default:
					lineToken = lineData.split("\\|", 2);
					if (lineToken.length > 1 &&
							signatures.containsSignature(lineToken[1]) == 0) {
						sig = lineToken[1];
						signatures.putSignature(sig);
						if (print) {
							int numCells = Signatures.signatureToNumberLiveCells(sig);

							System.out.println(lineToken[0] + "[" + numCells + "]|" + sig);
						}
					}
				}
			}
			if (findPeriodList != null) {
				findPeriod = toIntArray(findPeriodList);
			}
			if (findNCellsList != null) {
				findNCells = toIntArray(findNCellsList);
			}
			if (findExcludePeriodList != null) {
				findExcludePeriod = toIntArray(findExcludePeriodList);
			}
		} catch (IOException e) {
			//e.printStackTrace();
		}
	}

	public void generateDB(boolean print) {
		int[] data;
		String patternName;
		int ni = getNeighborIndex(), ri;
		Integer ruleIndex = getRuleIndex(ni);
		int start = ((print) ? -1 : 0); // decide about on screen
		final int MAX_RETRY = 16; // arbitrary... but do not want it looping forever

		generatingDB = true;
		signatures = new Signatures();
		findSignatures = new Signatures();
		// For application read from junk list
		if (neighbors != Life4DRules.indexToNeighbor[ni])
			System.out.println("NeighborIndex: " + ni + ", Neighbors: " +
				neighbors);
		if (ruleIndex == null) {
			//System.out.println("For neighbor index " + ni +
			//	", ruleIndex is null");
			return;
		}
		ri = ruleIndex;
		if (!print)
			clearField(visible);
		for (Life4DForms.Category c : Life4DForms.Category.values()) {
			for (int fi = start; fi < Life4DForms.length(ni, ri, c); fi++) {
				for (int retry = 0; retry <= MAX_RETRY; retry++) {
					if (fi == -1) { // do what is on screen
						patternName = this.name;
					} else {
						data = Life4DForms.get(ni, ri, c, fi);
						patternName = Life4DForms.getName(ni, ri, c, fi);
						if (data == null || data.length == 0) {
							System.out.println(fi + " data for ruleIndex " + ri + " is null");
							return;
						}
						displayForm(data, false);
					}
					// This fails under heavy load
					if (getSignatureList(print, true, (fi != -1 && c == Life4DForms.Category.OTHER), patternName)) {
						clearField(false);
						break;
					}
					clearField(false);
					if (retry == MAX_RETRY && wrap)
						System.out.println("FIXME");
					/*try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
					}*/
				}
			}
			start = 0;
		}
		addSignatures("found", print); // not needed if above works
		addSignatures("junk", print); // ignore these signatures
		if (print) {
			System.out.println("Number signatures: " +
				signatures.numberSignature());
			clearField(visible);
		}
		addSignatures("find", print);
		generatingDB = false;
	}

	// should randomize in W direction too but may be too confusing,
	// at least for me...
	private void setOrientCell(int polychoron, int x, int y, int z, int w,
			int startX, int startY, int startZ, int startW,
			byte state, int orient, boolean full) {
		int i = x, j = y, k = z, l = w, temp;

		if (full) {
			if (orient >= 72) {
				temp = i;
				i = k;
				k = j;
				j = l;
				l = temp;
			} else if (orient >= 48) {
				temp = i;
				i = k;
				k = l;
				l = j;
				j = temp;
			} else if (orient >= 32) {
				temp = i;
				i = j;
				j = k;
				k = l;
				l = temp;
			} else if (orient >= 16) {
				temp = i;
				i = l;
				l = k;
				k = j;
				j = temp;
			}
		} else {
			if (orient >= 32) {
				temp = i;
				i = j;
				j = k;
				k = temp;
			} else if (orient >= 16) {
				temp = i;
				i = k;
				k = j;
				j = temp;
			}
		}
		if (orient % 16 >= 8) {
			i = -i;
		}
		if (orient % 8 >= 4) {
			j = -j;
		}
		if (orient % 4 == 2) {
			k = -k;
		}
		if ((orient & 1) == 1) {
			l = -l;
		}
		i += startX;
		j += startY;
		k += startZ;
		l += startW;
		if (!cellCheck(polychoron, startX, startY, startZ, startW))

			k++;
		// do not draw cells here as they may be in wrong depth order
		if (i >= 0 && i < ncols && j >= 0 && j < nrows &&
				k >= 0 && k < nstacks &&
				l >= 0 && l < nzones)
			setCell(i, j, k, l, state);
	}

	public void displayForm(int[] data, boolean draw) {
		int n = 0, i = 0, j = 0, k = 0, l;
		int startX, startY, startZ, startW;
		int orient = 0;
		int polychoron = getPolychoron();

		startX = ncols / 2;
		startY = nrows / 2;
		startZ = nstacks / 2;
		startW = nzones / 2;
		if (!boundsCheck24(polychoron, startX, startY, startZ, startW))
			startZ++;
		if (randomOrient)
			orient = generator.nextInt(96);
		else if (randomOrient3D)
			orient = generator.nextInt(48);
		while (n < data.length) {
			switch (n % 4) {
			case 3:
			    l = data[n];
			    if (cellCheck(polychoron, i, j, k, l)) {
				if (randomOrient)
				    setOrientCell(polychoron, i, j, k, l,
					    startX, startY, startZ, startW,
					    (byte) 1, orient, true);
				else
				    setOrientCell(polychoron, i, j, k, l,
					    startX, startY, startZ, startW,
					    (byte) 1, orient, false);				}
			    break;
			case 2:
			    k = data[n];
			    break;
			case 1:
			    j = data[n];
			    break;
			default:
			    i = data[n];
			    break;
			}
			n++;
		}
		if (draw)
			redrawField();
	}

	public void getFormFromMenu(String string) {
		int[] data;
		int ni = getNeighborIndex(), ri;
		Integer ruleIndex = getRuleIndex(ni);

		name = string;
		if (neighbors != Life4DRules.indexToNeighbor[ni])
			System.out.println("NeighborIndex: " + ni + ", Neighbors: " +
				neighbors);
		Life4DForms lifeForms = new Life4DForms(ni);
		if (string.contains("Random")) {
			Integer symmetryIndex = Life4DSoup.stringToSymmetryIndex(string);

			if (symmetryIndex != null) {
				soup.displayRandomSoup(
					symmetryIndex,
					fillPercent, stirs,
					Life4DSoup.soupSize(ncols),
					Life4DSoup.soupSize(nrows),
					Life4DSoup.soupSize(nstacks),
					Life4DSoup.soupSize(nzones));
			}
			redrawField();
			return;
		}
		if (ruleIndex == null) {
			System.out.println(string + " ruleIndex is null for neighborIndex " + ni);
			return;
		}
		ri = ruleIndex;
		data = lifeForms.get(ni, ri, string);
		if (data == null || data.length == 0) {
			System.out.println(string + " data for ruleIndex " + ri + " is null");
			return;
		}
		displayForm(data, true);
	}

	public void getLife() {
		this.stringSave = ((Life4DFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getLifeString();
	}

	public void getLifeFile(String fileName, StringBuilder fileData) {
		try (BufferedReader br = new BufferedReader(
				new FileReader(fileName))) {
			String lineData;

			while ((lineData = br.readLine()) != null)
				fileData.append(lineData).append(NL);
			String out = fileData.toString();
			if (out == null || out.equals("")) {
				showMessage("Empty file to read");
			}
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getLifeFile() {
		StringBuilder fileData = new StringBuilder();
		getLifeFile(DATAFILE, fileData);
		this.stringSave = fileData.toString();
		getLifeString();
	}

	public void getLifeString() {
		getLifeString(this.stringSave, 0, 0, 0, 0);
	}

	public void getLifeString(String string,
			int x, int y, int z, int w) {
		try {
			String space = " ";
			String newline = NL;
			String pound = "#";
			int dim = 0, neighborhood, cells;
			int width = 5, height = 5, depth = 5, zone = 5;
			int offsetX = x, offsetY = y, offsetZ = z, offsetW = w;
			int i = 0;
			String ruleString;
			String[] lineToken = string.split(newline);
			if (lineToken.length == 0)
				return;
			String lineData = lineToken[i];
			String[] wordToken = null;

			try {
				while (i < lineToken.length) {
					wordToken = lineData.split(space);
					if (wordToken.length > 1 &&
							(pound + "Dim").equals(wordToken[0]))
						break;
					if (++i == lineToken.length)
						return;
					lineData = lineToken[i];
				}
				if (wordToken != null && wordToken.length > 1)
					dim = Integer.parseInt(wordToken[1]);
				if (dim != 4) {
					showMessage(DATAFILE +
						" corrupted: dim " + dim);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: dim not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Neighbors
				neighborhood = Integer.parseInt(wordToken[1]);
				if (Math.abs(neighborhood) <= 2) {
					showMessage(DATAFILE +
						" corrupted: neighbors " + neighborhood);
					neighborhood = this.neighbors;
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: neighbors not an integer");
				neighborhood = this.neighbors;
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #Rule
			ruleString = Life4DRules.parseRule(neighborhood, wordToken[1]);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Cells
				cells = Integer.parseInt(wordToken[1]);
				if (cells < 0) {
					showMessage(DATAFILE +
						" corrupted: cells " + cells);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: cells not an integer");
			}
			lineData = lineToken[++i];
			// #Name
			name = lineData.substring(lineData.indexOf(space) + 1);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Width
				width = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: width not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Height
				height = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: height not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Depth
				depth = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: depth not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Zone
				zone = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: depth not an integer");
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #P
			if (wordToken.length > 1)
				offsetX = Integer.parseInt(wordToken[1]);
			if (wordToken.length > 2)
				offsetY = Integer.parseInt(wordToken[2]);
			if (wordToken.length > 3)
				offsetZ = Integer.parseInt(wordToken[3]);
			if (wordToken.length > 4)
				offsetW = Integer.parseInt(wordToken[4]);
			//((Life4DFrame) frame).callback(
			//	Life4DInterface.ACTION_RESTORE);
			cells = scanField(lineToken, i, offsetX, offsetY, offsetZ, offsetW);
			drawField(true);
			System.out.println(DATAFILE +
				": neighbors " + neighborhood +
				", rule " + ruleString +
				", cells " + cells +
				", name " + name +
				", width " + width +
				", height " + height +
				", depth " + depth +
				", zone " + zone +
				", offsetX " + offsetX +
				", offsetY " + offsetY +
				", offsetZ " + offsetZ +
				", offsetW " + offsetW);
		} catch (NumberFormatException e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading");
		}
	}

	public void saveInitLife() {
		initSB.setLength(0);
		printField(initSB);
	}

	public void retrieveInitLife() {
		System.out.println(initSB);
	}

	public void writeLife() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		((Life4DFrame) frame).writeTextArea(sb.toString());
	}

	public void writeLifeFile() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.println(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	public boolean findPattern() {
		String newSignature = getGridSignature();

		if (newSignature != null) {
			if (findSignatures != null) {
				int mult = findSignatures.containsSignature(newSignature);

				if (mult != 0) {
					System.out.println("found (" + mult + ")!");
					System.out.println(newSignature);
					printLife();
					return (true);
				}
			}
			if (findNCells != null) {
				int numberCells = Signatures.signatureToNumberLiveCells(newSignature);

				for (int i = 0; i < findNCells.length; i++) {
					if (findNCells[i] == numberCells && (signatures == null ||
							signatures.containsSignature(newSignature) == 0)) {
						checkFind = i; //findPeriod[i];
						 //System.out.println("found: " + i + " " + findPeriod[i]);
					}
				}
			}
		}
		return (signatures != null && signatures.containsSignature(newSignature) != 0);
	}

	public synchronized boolean printSearch(int iter) {
		if (iter == 0) {
			stamp = getGridStamp();
			signature = getGridSignature();
			oldnCells = nCells = Signatures.signatureToNumberLiveCells(signature);
			//System.out.println(signature + " in db");
			return (signatures == null || signatures.containsSignature(signature) == 0);
		}

		String newStamp = getGridStamp();
		String newSignature = getGridSignature();
		nCells = Signatures.signatureToNumberLiveCells(newSignature);
		int blob = (int) ((1.0 - (double) cCells / (2 * nCells)) * 100.0);

		if ((oldnCells <= searchCount || nCells <= searchCount) &&
				blob <= blobPercent &&
				((findStable && cCells == 0 && nCells != 0 &&
				newStamp.equals(stamp) &&
				newSignature.equals(signature)) ||
				/*!findStable &&*/
				(oldnCells != nCells ||
				!newStamp.equals(stamp) ||
				!newSignature.equals(signature))) &&
				(excludeCount < 2 ||
				oldnCells % excludeCount != 0 ||
				nCells % excludeCount != 0)) {
			if (signatures != null && signatures.containsSignature(newSignature) != 0) {
				return false;
			}
			if (stamp != null)
				System.out.println(stamp);
			if (newStamp != null)
				System.out.println(newStamp);
			if (signature != null &&
					!signature.equals(newSignature)) {
				System.out.println(signature);
				if (signatures != null)
					signatures.putSignature(signature);
			}
			if (newSignature != null) {
				System.out.println(newSignature);
				if (signatures != null)
					signatures.putSignature(newSignature);
			}
			if (genOffset > 0)
				System.out.println("Edge hit");
			printLife();
		}
		return true;
	}

	public void printLife() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		System.out.println(sb);
	}

	void eraseFrame(Graphics g) {
		g.setColor(background);
		g.fillRect(0, 0, getSize().width, getSize().height);
	}

	void eraseFrame() {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			eraseFrame(g);
		} finally {
			g.dispose();
		}
	}

	void draw4DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	/*void drawFrame(Graphics g, boolean show) {
		Box4D b = cellPl.boundBox(ncols, nrows, nstacks);

		if (g == null || b == null)
			return;
		g.setColor(background.darker().darker());
		g.fillRect(0, 0, getSize().width, b.y);
		g.fillRect(0, b.y, b.x, b.height - b.y);
		g.fillRect(0, b.height, getSize().width, b.y);
		g.fillRect(b.width, b.y,
			getSize().width - b.width, b.height - b.y);
		draw4DFrame(show && focus, b.x, b.y,
			b.width - b.x, b.height - b.y);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}*/

	void recalibrateBounds(int width, int height, int depth, int shift) {
		boolean first = true;
		minCol = width / 2 - 1;
		maxCol = width / 2 + 1;
		minRow = height / 2 - 1;
		maxRow = height / 2 + 1;
		minStack = depth / 2 - 1;
		maxStack = depth / 2 + 1;
		minZone = shift / 2 - 1;
		maxZone = shift / 2 + 1;

		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int areaBounds = width * height;
			int volumeBounds = areaBounds * depth;
			int col = getColumn(position, width);
			int row = getRow(position, width, areaBounds);
			int stack = getStack(position, areaBounds, volumeBounds);
			int zone = getZone(position, volumeBounds);
			if (first) {
				minCol = maxCol = col;
				if (col > 0)
					minCol--;
				if (col < width - 1)
					maxCol++;
				minRow = maxRow = row;
				if (row > 0)
					minRow--;
				if (row < height - 1)
					maxRow++;
				minStack = maxStack = stack;
				if (stack > 0)
					minStack--;
				if (stack < depth - 1)
					maxStack++;
				minZone = maxZone = zone;
				if (shift > 0)
					minStack--;
				if (zone < shift - 1)
					maxStack++;
				first = false;
			} else {
				if (col - 1 < minCol && col > 0)
					minCol = col - 1;
				if (col + 1 > maxCol && col < width - 1)
					maxCol = col + 1;
				if (row - 1 < minRow && row > 0)
					minRow = row - 1;
				if (row + 1 > maxRow && row < height - 1)
					maxRow = row + 1;
				if (stack - 1 < minStack && stack > 0)
					minStack = stack - 1;
				if (stack + 1 > maxStack && stack < depth - 1)
					maxStack = stack + 1;
				if (zone - 1 < minZone && shift > 0)
					minZone = zone - 1;
				if (zone + 1 > maxStack && zone < shift - 1)
					maxZone = zone + 1;
			}
		}
	}

	boolean checkAliveBounds(int i, int j, int k, int l) {
		int offset = (neighbors == -32) ? 1 : 0;

		if (minCol <= offset || maxCol >= ncols - 1 - offset ||
				i <= offset || i >= ncols - 1 - offset) {
			minCol = 0;
			maxCol = ncols - 1;
			edge = true;
		} else if (i <= minCol) {
			minCol = i - 1;
		} else if (i >= maxCol) {
			maxCol = i + 1;
		}
		if (minRow <= offset || maxRow >= nrows - 1 - offset ||
				j <= offset || j >= nrows - 1 - offset) {
			minRow = 0;
			maxRow = nrows - 1;
			edge = true;
		} else if (j <= minRow) {
			minRow = j - 1;
		} else if (j >= maxRow) {
			maxRow = j + 1;
		}
		if (minStack <= offset || maxStack >= nstacks - 1 - offset ||
				k <= offset || k >= nstacks - 1 - offset) {
			minStack = 0;
			maxStack = nstacks - 1;
			edge = true;
		} else if (k <= minStack) {
			minStack = k - 1;
		} else if (k >= maxStack) {
			maxStack = k + 1;
		}
		if (minZone <= offset || maxZone >= nzones - 1 - offset ||
				l <= offset || l >= nzones - 1 - offset) {
			minZone = 0;
			maxZone = nzones - 1;
			edge = true;
		} else if (l <= minZone) {
			minZone = l - 1;
		} else if (l >= maxZone) {
			maxZone = l + 1;
		}
		return edge;
	}

	int scanField(String[] lineToken, int stringIndex,
			int offsetX, int offsetY, int offsetZ, int offsetW) {
		int polychoron = getPolychoron();
		String lineData;
		int m = stringIndex, count = 0;
		int i, j, k, l, n, p, startX, startY, startZ, startW;
		String space = " ";
		int numNL = 0;

		startX = offsetX + ncols / 2;
		startY = offsetY + nrows / 2;
		startZ = offsetZ + nstacks / 2;
		startW = offsetW + nzones / 2;
		if (!boundsCheck24(polychoron, startX, startY, startZ, startW))
			startZ++;
		try {
			l = startW;
			k = startZ;
			for (j = startY; ++m < lineToken.length; j++) {
				lineData = lineToken[m];
				if (lineData.length() == 0) {
					if (numNL > 1)
						break; // allows flexibility
					if (numNL == 1) {
						l++;
						k = startZ;
					} else {
						k++;
					}
					j = startY - 1;
					numNL++;
					continue;
				}
				numNL = 0;
				n = 0;
				i = startX;
				while (n < lineData.length()) {
					byte oldCell, newCell;

					if (space.charAt(0) == lineData.charAt(n)) {
						n++;
						continue;
					}
					/*if (i < 0 || i >= ncols ||
							j < 0 || j >= nrows ||
							k < 0 || k >= nstacks ||
							l < 0 || l >= nzones) {
						n++;
						continue;
					}*/
					p = getPosition(i, j, k, l);
					newCell = 0;
					if (cellCheck(polychoron, i, j, k, l)) {
						if (states == 2) {
							newCell = (byte) ((Life4DInterface.aliveChar(lineData.charAt(n))) ? 1 : 0);
						} else {
							if (!Life4DInterface.deadChar(lineData.charAt(n))) {
								newCell = (byte) (lineData.charAt(n) - '0');
							}
						}
					}
					if (newCell != 0) {
						count++;
					}
					oldCell = field.getCell(p);
					if (oldCell != newCell) {
						field.setCell(p, newCell);
						/*CellInfo info = new CellInfo(
							p, newCell);
						drawCell(info);*/
						if (newCell == 0) {
							nCells--;
						} else {
							nCells++;
							checkAliveBounds(i, j, k, l);
						}
						((Life4DFrame) frame).updateCells(nCells);
					}
					i++;
					n++;
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading (scanField)");
		}
		return count;
	}

	public synchronized void drawField(boolean draw) {
		nCells = cellCount();
		if (draw) {
			showLiveCells();
			if (grid)
				drawGrid();
			((Life4DFrame) frame).updateCells(nCells);
			capturer.show();
		}
	}

	void redrawField() {
		eraseFrame();
		//drawFrame(true);
		drawField(true);
	}

	public Box4D getSelectedBox() {
		Point4D p = new Point4D();
		Dimension4D d = new Dimension4D();
		Rectangle rect = capturer.getRectangle();

		if (rect == null)
			return null;
		p.x = rect.x;
		p.y = rect.y;
		d.width = rect.width;
		d.height = rect.height;
		p.z = 0; // TODO: need a way to select in 4D
		d.depth = 1; // TODO: need a way to select in 4D
		p.w = 0; // TODO: need a way to select in 4D
		d.zone = 1; //TODO: need a way to select in 4D
		return new Box4D(p, d);
	}

	void banner(String msg) {
		((Life4DFrame) frame).displayMessage(msg);
	}

	Point4D inGrid(Point pt) {
		Graphics g = this.getGraphics();
		int k = mouseDepth;
		int l = mouseZone;
		int polychoron = getPolychoron();

		if (g == null)
			return new Point4D(-1, -1, -1, -1);
		for (int j = 0; j < nrows; j++)
		for (int i = 0; i < ncols; i++)
		if (cellCheck(polychoron, i, j, k, l) &&
				cellPl.inCell(g4d, pt, i, j, k/*, l*/)) {
			//System.out.println(i + " " + j + " " + k + " " + l);
			return new Point4D(i, j, k, l);
		}
		return new Point4D(-1, -1, -1, -1);
	}

	void drawGrid() {
		Graphics g = this.getGraphics();
		Color color;
		int k = mouseDepth;
		int l = mouseZone;
		int polychoron = getPolychoron();

		if (g == null)
			return;
		if (grid)
			color = Color.black;
		else
			color = background;
		g.setColor(color);
		for (int j = 0; j < nrows; j++)
		for (int i = 0; i < ncols; i++)
		if (cellCheck(polychoron, i, j, k, l)) {
			cellPl.drawCell(g4d, false, wireframe, null, null, i, j, k/*, l*/);
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		//drawFrame(focus);
		drawField(true);
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}

	int getColumn(int position) {
		return position % this.ncols;
	}

	int getRow(int position) {
		return (position % this.area) / this.ncols;
	}

	int getStack(int position) {
		return (position % this.volume) / this.area;
	}

	int getZone(int position) {
		return position / this.volume;
	}

	static int getColumn(int position, int ncols) {
		return position % ncols;
	}

	static int getRow(int position, int ncols, int area) {
		return (position % area) / ncols;
	}

	static int getStack(int position, int area, int volume) {
		return (position % volume) / area;
	}

	static int getZone(int position, int volume) {
		return position / volume;
	}

	int getPosition(int col, int row, int stack, int zone) {
		return getPosition(col, row, stack, zone,
			 this.ncols, this.area, this.volume);
	}

	static int getPosition(int col, int row, int stack, int zone,
			int ncols, int area, int volume) {
		return zone * volume + stack * area + row * ncols + col;
	}

	int getColumn(CellInfo info) {
		return getColumn(info.getPosition());
	}

	int getRow(CellInfo info) {
		return getRow(info.getPosition());
	}

	int getStack(CellInfo info) {
		return getStack(info.getPosition());
	}

	int getZone(CellInfo info) {
		return getZone(info.getPosition());
	}

	void setPosition(CellInfo info, int col, int row, int stack, int zone) {
		info.setPosition(getPosition(col, row, stack, zone));
	}

	int getNeighbors(int col, int row, int stack, int zone) {
		int n, p, count = 0;

		if (col == 0 || row == 0 || stack == 0 || zone == 0 ||
				col == ncols - 1 || row == nrows - 1 ||
				stack == nstacks - 1 || zone == nzones - 1) {
			edge = true;
			if (!wrap)
				return 0;
		}
		for (n = 0; n < Math.abs(neighbors); n++) {
			p = positionOfNeighbor(n, col, row, stack, zone);
			try {
				int state = field.getCell(p);

				if (states == 2) {
					if (state != 0) {
						count++; // totalistic
					}
				} else { // if (states == 4) // double cell
					if (state == 1 || state == 2) {
						count++; // totalistic
					} else if (state == 3) {
						count += 2;
					}
				}
			} catch (ArrayIndexOutOfBoundsException e) {
				e.printStackTrace();
				System.out.println(n + " " + p + ": " +
					col + " " + row + " " + stack + " " + zone);
			}
		}
		return count;
	}

	int positionOfNeighbor(int cell, int i, int j, int k, int l) {
		int col = i, row = j, stack = k, zone = l, n = cell;

		// 8 faces
		// 16 corners
		// 24 face_sides
		// -24 (alt 24) faces + corners
		// 32 side_corners
		// not implemented (alt way) faces + face_sides
		// -32 (another alt 32) face_sides + far corners
		// 40 faces + side_corners
		// -40 (alt 40) corners + face_sides
		// 48 exclude side_corners
		// -48 (alt 48) corners + side_corners
		// 56 exclude face_sides
		// -56 (alt 56) face_sides + side_corners
		// 64 exclude corners
		// 72 exclude faces
		// 80 all
		/* need to add a flag if want alternate representation */
		/* corners */
		if (neighbors == 8 || /*neighbors == -32 ||*/
				neighbors == 40 || neighbors == 48 ||
				neighbors == 56 || neighbors == 64 ||
				neighbors == 80) {
			switch (n) {
			case 0:
				col = (col == 0) ? ncols - 1 : col - 1;
				break;
			case 1:
				col = (col + 1 == ncols) ? 0 : col + 1;
				break;
			case 2:
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 3:
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 4:
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 5:
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 6:
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 7:
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			default:
				break;
			}
		} else {
			n += 8;
		}
		/* face_sides */
		if (neighbors == 24 || neighbors == -32 ||
				neighbors == -40 || neighbors == 48 ||
				neighbors == -56 || neighbors == 64 ||
				neighbors == 72 || neighbors == 80) {
			switch (n) {
			case 8:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 9:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 10:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 11:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 12:
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 13:
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 14:
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 15:
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 16:
				col = (col == 0) ? ncols - 1 : col - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 17:
				col = (col + 1 == ncols) ? 0 : col + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 18:
				col = (col == 0) ? ncols - 1 : col - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 19:
				col = (col + 1 == ncols) ? 0 : col + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 20:
				row = (row == 0) ? nrows - 1 : row - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 21:
				row = (row + 1 == nrows) ? 0 : row + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 22:
				row = (row == 0) ? nrows - 1 : row - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 23:
				row = (row + 1 == nrows) ? 0 : row + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 24:
				col = (col == 0) ? ncols - 1 : col - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 25:
				col = (col + 1 == ncols) ? 0 : col + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 26:
				col = (col == 0) ? ncols - 1 : col - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 27:
				col = (col + 1 == ncols) ? 0 : col + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 28:
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 29:
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 30:
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 31:
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			default:
				break;
			}
		} else {
			n += 24;
		}
		/* side_corners */
		if (neighbors == 32 || neighbors == 40 ||
				neighbors == -48 || neighbors == 56 ||
				neighbors == -56 || neighbors == 64 ||
				neighbors == 72 || neighbors == 80) {
			switch (n) {
			case 32:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 33:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 34:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 35:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				break;
			case 36:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 37:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 38:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 39:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				break;
			case 40:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 41:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 42:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 43:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 44:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 45:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 46:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 47:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 48:
				col = (col == 0) ? ncols - 1 : col - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 49:
				col = (col + 1 == ncols) ? 0 : col + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 50:
				col = (col == 0) ? ncols - 1 : col - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 51:
				col = (col + 1 == ncols) ? 0 : col + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 52:
				col = (col == 0) ? ncols - 1 : col - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 53:
				col = (col + 1 == ncols) ? 0 : col + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 54:
				col = (col == 0) ? ncols - 1 : col - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 55:
				col = (col + 1 == ncols) ? 0 : col + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 56:
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 57:
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 58:
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 59:
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 60:
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 61:
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 62:
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 63:
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			default:
				break;
			}
		} else {
			n += 32;
		}
		/* corners */
		if (neighbors == 16 || neighbors == -24 ||
				neighbors == -40 || neighbors == 48 ||
				neighbors == -48 || neighbors == 56 ||
				neighbors == 72 || neighbors == 80) {
			switch (n) {
			case 64:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 65:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 66:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 67:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 68:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 69:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 70:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 71:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone == 0) ? nzones - 1 : zone - 1;
				break;
			case 72:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 73:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 74:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 75:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack == 0) ? nstacks - 1 : stack - 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 76:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 77:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 78:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			case 79:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				stack = (stack + 1 == nstacks) ? 0 : stack + 1;
				zone = (zone + 1 == nzones) ? 0 : zone + 1;
				break;
			default:
				break;
			}
		} else {
			n += 16;
		}
		/* corners */
		if (neighbors == -32) {
			switch (n) {
			case 80:
				col = (col < 2) ? ncols - 2 + col : col - 2;
				break;
			case 81:
				col = (col + 2 >= ncols) ? col + 2 - ncols : col + 2;
				break;
			case 82:
				row = (row < 2) ? nrows - 2 + row : row - 2;
				break;
			case 83:
				row = (row + 2 >= nrows) ? row + 2 - nrows : row + 2;
				break;
			case 84:
				stack = (stack < 2) ? nstacks - 2 + stack : stack - 2;
				break;
			case 85:
				stack = (stack + 2 >= nstacks) ? stack + 2 - nstacks : stack + 2;
				break;
			case 86:
				zone = (zone < 2) ? nzones - 2 + zone : zone - 2;
				break;
			case 87:
				zone = (zone + 2 >= nzones) ? zone + 2 - nzones : zone + 2;
				break;
			default:
				break;
			}
		}
		return (getPosition(col, row, stack, zone));
	}

	/* O(n) operation for a cell but only done for mouse click. */
	byte incCell(int col, int row, int stack, int zone) {
		int p = getPosition(col, row, stack, zone);
		byte state = field.getCell(p);
		byte nextState = (byte) ((state < states - 1) ?
			state + 1 : 0);

		if (state == 0) {
			field.setCell(p, nextState);
		} else {
			for (Iterator<CellInfo> iterator =
					field.list.iterator();
					iterator.hasNext();) {
				CellInfo cellInfo = iterator.next();
				int position = cellInfo.getPosition();
				if (position == p) {
					field.setArrayCell(position, nextState);
					if (nextState == 0) {
						iterator.remove();
					} else {
						cellInfo.setCell(nextState);
					}
					break;
				}
			}
		}
		if (state == 0) {
			nCells++;
			checkAliveBounds(col, row, stack, zone);
		} else if (state == states - 1) {
			nCells--;
		}
		//CellInfo info = new CellInfo(p, nextState);
		/*if (nextState == 0) {
			eraseCell(info);
		} else {
			drawCell(info);
		}*/
		((Life4DFrame) frame).updateCells(nCells);
		return nextState;
	}

	int cellCount() {
		return field.list.size();
	}

	boolean setCell(int col, int row, int stack, int zone,
			byte nextState) {
		int polychoron = getPolychoron();
		int p = getPosition(col, row, stack, zone);
		byte state = field.getCell(p);

		if (state == 0 && nextState != 0) {
			field.setCell(p, nextState);
			if (!cellCheck(polychoron, col, row, stack, zone))
				throw new IllegalArgumentException("failed cell check on draw");
			nCells++;
			checkAliveBounds(col, row, stack, zone);
		} else if (state == nextState) {
			return false;
		} else if (nextState == 0) { // state != 0
			field.setCell(p, (byte) 0);
			nCells--;
			/*CellInfo info = new CellInfo(p, nextState);*/
			/*eraseCell(info);*/
			return false;
		} else { // different sort of live cell
			field.setCell(p, nextState);
			cCells++;
		}
		/*if (draw) {
			CellInfo info = new CellInfo(p, (byte) 0,
				nextState, false);

			drawCell(info);
			//showLiveCells();
			((Life4DFrame) frame).updateCells(nCells);
		}*/
		return true;
	}

	/*void printState(int state) {
		CellList curr;
		int i = 0;

		curr = first[state]->next;
		System.out.println("state " + state);
		while (curr != list[state]) {
			System.out.println(i + ": position " +
				curr.info.getPosition() + ",	age " +
				curr.info.getAge() + ", state " +
				curr.info.getState() + ", toggle " +
				curr.info.getToggle());
			curr = curr.next;
			i++;
		}
	}*/

	static Color colorRainbow(int state, int factor) {
		final int MAX_INTENSITY = 255;

		if (state >= 0 && state < factor) {
			return new Color(MAX_INTENSITY, state * 256 / factor, 0);
		} else if (state >= factor && state < factor * 2) {
			return new Color(511 - state * 256 / factor, MAX_INTENSITY, 0);
		} else if (state >= factor * 2 && state < factor * 3) {
			return new Color(0, MAX_INTENSITY, state * 256 / factor - 512);
		} else if (state >= factor * 3 && state < factor * 4 - 1) { /* slight defect */
			return new Color(0, 1023 - state * 256 / factor, MAX_INTENSITY);
		} else {
			return new Color(0, 0, MAX_INTENSITY);
		}
	}

	void drawCell(CellInfo info) {
		int zone = getZone(info);
		if (zone != mouseZone)
			return;
		Color color;
		int polychoron = getPolychoron();
		int col = getColumn(info);
		int row = getRow(info);
		int stack = getStack(info);
		byte cell = info.getCell();

		if (!cellCheck(polychoron, col, row, stack, zone))
			return;
		if (mono) {
			color = Color.white;
		} else if (states == 2) {
			color = colorRainbow(cell, 16);
		} else {
			color = colorRainbow((cell - 1) * 16, 16);
		}
		if (g4d == null)
			return;
		if (g4d.getGraphics() == null) {
			/* can happen on static field */
			resizeField();
		}
		if (!color.equals(background)) // Changed for 3D
			cellPl.drawCell(g4d, round, wireframe,
				color, gridColor, col, row, stack/*, zone*/);
	}

	void eraseCell(CellInfo info) {
		int zone = getZone(info);
		if (zone != mouseZone)
			return;
		//Graphics g = this.getGraphics();
		int col = getColumn(info);
		int row = getRow(info);
		int stack = getStack(info);

		cellPl.drawCell(g4d, false, wireframe, background, gridColor,
			col, row, stack/*, zone*/);
	}

	Life4DMatrix getField() {
		return field;
	}

class Capturer {
	private Point4D anchor = new Point4D(0, 0, 0, 0);
	private Point4D movable = new Point4D(0, 0, 0, 0);
	private Point4D start = new Point4D(0, 0, 0, 0);
	private Point4D stop = new Point4D(0, 0, 0, 0);
	private int completed = -1;
	private final Color selectedColor = Color.magenta;

	Point4D calcPoint(int x, int y, int z, int w) {
		Point4D t = new Point4D(x, y, z, w);

		if (t.x < 0)
			t.x = 0;
		if (t.y < 0)
			t.y = 0;
		if (t.z < 0)
			t.z = 0;
		if (t.w < 0)
			t.w = 0;
		if (t.x >= ncols)
			t.x = ncols - 1;
		if (t.y >= nrows)
			t.y = nrows - 1;
		if (t.z >= nstacks)
			t.z = nstacks - 1;
		if (t.w >= nzones)
			t.w = nzones - 1;
		return t;
	}

	void calcStartPoint() {
		if (anchor.x > movable.x)
			start.x = movable.x;
		else
			start.x = anchor.x;
		if (anchor.y > movable.y)
			start.y = movable.y;
		else
			start.y = anchor.y;
		if (anchor.z > movable.z)
			start.z = movable.z;
		else
			start.z = anchor.z;
	}

	void calcEndPoint() {
		if (anchor.x < movable.x)
			stop.x = movable.x;
		else
			stop.x = anchor.x;
		if (anchor.y < movable.y)
			stop.y = movable.y;
		else
			stop.y = anchor.y;
		if (anchor.z < movable.z)
			stop.z = movable.z;
		else
			stop.z = anchor.z;
	}

	void startCapture(int column, int row, int stack, int zone) {
		movable = anchor = calcPoint(column, row, stack, zone);
		calcStartPoint();
		calcEndPoint();
		completed = 0;
		drawCaptureBox((grid) ? gridColor : background);
	}

	void continueCapture(int column, int row, int stack, int zone) {
		int captureColumns = getColumns();
		int captureRows = getRows();

		movable = calcPoint(column, row, stack, zone);
		drawCaptureBox((grid) ? gridColor : background);
		if (column >= 0 && row >= 0 && stack >= 0 && zone >= 0 &&
			column < captureColumns && row < captureRows && stack < stacks && zone < zones) {
			calcStartPoint();
			calcEndPoint();
			drawCaptureBox(selectedColor);
			return;
		}
		calcStartPoint();
		calcEndPoint();
	}

	void stopCapture() {
		if (completed == 0)
			completed = 1;
	}

	void clearCapture() {
		drawCaptureBox((grid) ? gridColor : background);
		anchor.setLocation(0, 0);
		movable.setLocation(0, 0);
		completed = -1;
	}

	private void drawCaptureBox(Color color) {
		int x = start.x;
		int y = start.y;
		int z = start.z;
		int w = start.w;
		int width = stop.x - start.x + 1;
		int height = stop.y - start.y + 1;
		int depth = stop.z - start.z + 1;
		int zone = stop.w - start.w + 1;

		if (completed == -1)
			return;
		cellPl.drawCaptureBox(g4d, color,
			x, y, z, w, width, height, depth, zone);
	}

	void show() {
		if (!anchor.equals(movable))
			drawCaptureBox(selectedColor);
	}

	Rectangle getRectangle() {
		if (completed == 1) {
			int x = start.x;
			int y = start.y;
			int width = stop.x - start.x + 1;
			int height = stop.y - start.y + 1;

			return new Rectangle(x, y, width, height);
		}
		return null;
	}

}

}
