package com.sillycycle.bagleyd.life4d;

/* Code for selection spliced in is due to Alex Shvets */
/* http://home.comcast.net/~shvets/Programs/Examples/GameLife/life2.html */
// Life4DPopup.java

import java.awt.Cursor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;

import com.sillycycle.bagleyd.util.Box4D;

public class Life4DPopup implements ClipboardOwner, ActionListener {
	Clipboard clipboard = new Clipboard("Life");
	Life4DCanvas life4d;
	static final boolean DEBUG = false;

	public Life4DPopup(Life4DCanvas life4d) {
		this.life4d = life4d;
	}

	@Override
	public void lostOwnership(Clipboard lostClipboard, Transferable contents) {
		// implements ClipboardOwner
	}

	// implements ActionListener
	@Override
	public void actionPerformed(ActionEvent e) {
		String command = e.getActionCommand();
		Box4D selectedBox = getSelectedBox();
		Life4DMatrix field = life4d.getField();

		if (DEBUG) {
			if (selectedBox == null)
				System.out.println("actionPeformed: selectedBox = null");
			else if (life4d.getField() == null)
				System.out.println("actionPeformed: field = null");
		}
		if (selectedBox == null || field == null)
			return;
		if (DEBUG) {
			System.out.println("actionPeformed: " + command);
		}
		life4d.setCursor(new Cursor(Cursor.WAIT_CURSOR));
		switch(command) {
		case "Cut":
			doCut(selectedBox);
			break;
		case "Copy":
			doCopy(selectedBox);
			break;
		case "Paste":
			doPaste(selectedBox);
			break;
		case "Clear":
			doClear(selectedBox);
			break;
		case "Cw":
			doRotate(selectedBox, 0, true);
			break;
		case "Ccw":
			doRotate(selectedBox, 0, false);
			break;
		case "Refl Wall":
			doReflect(selectedBox, 0, true);
			break;
		case "Refl Grnd":
			doReflect(selectedBox, 0, false);
			break;
		default:
		}
		life4d.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	}

	public Box4D getSelectedBox() {
		return life4d.getSelectedBox();
	}

	public void doClear(Box4D selectedBox) {
		Life4DMatrix field = life4d.getField();

		field.clearSubMatrix(selectedBox);
		life4d.redrawField(); //(selectedBox);
	}

	public void doCopy(Box4D selectedBox) {
		Life4DMatrix field = life4d.getField();

		Life4DMatrix selection = field.getSubMatrix(selectedBox);
		Life4DSelection data = new Life4DSelection(selection);
		clipboard.setContents(data, this);
	}

	public void doCut(Box4D selectedBox) {
		Life4DMatrix field = life4d.getField();
		Life4DMatrix selection = field.getSubMatrix(selectedBox);
		Life4DSelection data = new Life4DSelection(selection);

		clipboard.setContents(data, this);

		field.clearSubMatrix(selectedBox);
		life4d.redrawField(); //(selectedBox);
	}

	public void doPaste(Box4D selectedBox) {
		Life4DMatrix field = life4d.getField();

		try {
			Transferable data = clipboard.getContents(this);
			if (data == null) {
				return;
			}
			Life4DMatrix selection = (Life4DMatrix) data.getTransferData(
				Life4DSelection.lifeFlavor);

			selectedBox.setSize(selection.getColumns(),
				selection.getRows());
			field.setSubMatrix(selection, selectedBox);
			life4d.redrawField(); //(selectedBox);
		} catch (UnsupportedFlavorException | IOException e) {
			e.printStackTrace();
		}
	}

	public void doRotate(Box4D selectedBox, int plane, boolean cw) {
		Life4DMatrix field = life4d.getField();

		try {
			Transferable data = clipboard.getContents(this);
			if (data == null) {
				return;
			}
			Life4DMatrix selection = (Life4DMatrix) data.getTransferData(
				Life4DSelection.lifeFlavor);

			selectedBox.setSize(selection.getColumns(),
				selection.getRows());
			field.rotate(selection, selectedBox,
				plane, cw);
			life4d.redrawField(); //(selectedBox);
		} catch (UnsupportedFlavorException | IOException e) {
			e.printStackTrace();
		}
	}

	public void doReflect(Box4D selectedBox, int plane, boolean wall) {
		Life4DMatrix field = life4d.getField();

		try {
			Transferable data = clipboard.getContents(this);
			if (data == null) {
				return;
			}
			Life4DMatrix selection = (Life4DMatrix) data.getTransferData(
				Life4DSelection.lifeFlavor);

			selectedBox.setSize(selection.getColumns(),
				selection.getRows());
			field.reflect(selection, selectedBox,
				plane, wall);
			life4d.redrawField(); //(selectedBox);
		} catch (UnsupportedFlavorException | IOException e) {
			e.printStackTrace();
		}
	}
}
