package com.sillycycle.bagleyd.life4d;

/*
 * @(#)Life4DSoup.java
 *
 * Copyright 2010 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;

/**
 * The <code>Life4DSoup</code> class does all the graphics for
 * <code>Life4DCanvas</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Life4DSoup {
	static Life4DCanvas canvas = null;
	Random generator;

	static final String[] symmetryStrings = {
		"Full Random",
		"Sym X Random",
		"Sym Y Random",
		"Sym XY Random",
		"Antisym X Random",
		"Antisym Y Random",
		"Half Antisym Random", // Diagonal
		"Half Sym Random", // Mirror Diagonal
		"Full Antisym Random", // Square
		"Full Sym Random", // Mirror Square
		"Half Antisym Skew Random",
		"Sym XYZ Random",
		"Half Sym Skew Random",
		"Full Antisym Cube Random",
		"Full Sym Cube Random",
		"Full Tetra Random",
		"Alt Full Antisym Cube Random",
		"Antisym Z Random",
		"Half Antisym Metaskew Random",
		/* Non Glider Patterns */
		"Sym XYZW Random",
		"Half Sym Metaskew Random",
		"Full Antisym Tesseract Random",
		"Full Sym Tesseract Random",
		"Full Penta Random",
		"Antisym W Random",
	};

	public Life4DSoup(Life4DCanvas canvas) {
		Life4DSoup.canvas = canvas;
		generator = new Random(System.nanoTime());
	}


	static Integer stringToSymmetryIndex(String string) {
		for (int i = 0; i < symmetryStrings.length; i++) {
			if (symmetryStrings[i].equals(string)) {
				return i;
			}
		}
		return null;
	}

	static int soupSize(int s) {
		int val = s / 12;
		final int bound = 6;
		return ((val > bound) ? val : bound);
	}

	int randomSymmetry(boolean gliderSearch) {
		// For glider searches, too much symmetry can cause the
		// gliders to be short lived since wrap is used to find them.
		// For pattern searches, full random, i.e. no symmetry
		// is excluded but these patterns can still show up in
		// symmetrical pairs, etc.
		final int vector = 19;

		if (gliderSearch) {
			return generator.nextInt(vector);  /* all 3D ones + Metaskew Diagonal */
		}
		//return generator.nextInt(symmetryStrings.length - 1) + 1;
		return generator.nextInt(symmetryStrings.length);
		//return generator.nextInt(symmetryStrings.length - vector) + vector;
	}

	public void displayRandomSoup(int symmetryIndex, int percentChance, int stirs,
			int fieldX, int fieldY, int fieldZ, int fieldW) {
		int polychoron = canvas.getPolychoron();
		int ncols = canvas.getColumns();
		int nrows = canvas.getRows();
		int nstacks = canvas.getStacks();
		int nzones = canvas.getZones();
		int col, row, stack, zone;
		int halfcols = ncols / 2;
		int halfrows = nrows / 2;
		int halfstacks = nstacks / 2;
		int halfzones = nzones / 2;
		int vx = fieldX / 2 + 1;
		int vy = fieldY / 2 + 1;
		int vz = fieldZ / 2 + 1;
		int vw = fieldW / 2 + 1;
		int even = 0;
		int yEven = 0;
		int zEven = 0;
		int wEven = 0;
		int state = 1;
		int percent = percentChance;

		if (vx < 1)
			vx = 1;
		if (vy < 1)
			vy = 1;
		if (vz < 1)
			vz = 1;
		if (vw < 1)
			vw = 1;
		if (stirs > 0) {
			/* make smaller as it should expand */
			vx = (fieldX / 3) | 1;
			vy = (fieldY / 3) | 1;
			vz = (fieldZ / 3) | 1;
			vw = (fieldW / 3) | 1;
		}
		if (polychoron != 24) {
			even = generator.nextInt(2);
			yEven = generator.nextInt(2);
			zEven = generator.nextInt(2);
			wEven = generator.nextInt(2);
		}
		//System.out.println("symmetryIndex = " + symmetryIndex);
		switch (symmetryIndex) {
		case 0: /* Full Random */
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = halfrows - vy; row < halfrows + vy; ++row)
			for (col = halfcols - vx; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
			}
			break;
		case 1: /* Sym X Random */
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = halfrows - vy; row < halfrows + vy; ++row)
			for (col = halfcols; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, row, stack, zone, state);
			}
			break;
		case 2: /* Sym Y Random */
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = halfrows; row < halfrows + vy; ++row)
			for (col = halfcols - vx; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
				canvas.setCell(col, 2 * halfrows - row - even, stack, zone, state);
			}
			break;
		case 3: /* Sym XY Random */
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = halfrows; row < halfrows + vy; ++row)
			for (col = halfcols; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
				canvas.setCell(col, 2 * halfrows - row - yEven, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, row, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, stack, zone, state);
			}
			break;
		case 4: /* Antisym X Random */
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = halfrows - vy; row < halfrows + vy; ++row)
			for (col = halfcols; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, stack, zone, state);
			}
			break;
		case 5: /* Antisym Y Random */
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = halfrows; row < halfrows + vy; ++row)
			for (col = halfcols - vx; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, stack, zone, state);
			}
			break;
		case 6: /* Half Antisym Random */
			// Diagonal
			vx = vy = (vx + vy) / 2;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = -vy; row < vy; ++row)
			for (col = row; col < vx; ++col)
			if (even == 1) {
				if (Life4DCanvas.cellCheck(polychoron, col + halfcols, row + halfrows, stack, zone)
						&& generator.nextInt(100) < percent) {
					canvas.setCell(col + halfcols, row + halfrows, stack, zone, state);
					canvas.setCell(row + halfcols, col + halfrows, stack, zone, state);
				}
			} else {
				if (Life4DCanvas.cellCheck(polychoron, -col + halfcols, row + halfrows, stack, zone)
						&& generator.nextInt(100) < percent) {
					canvas.setCell(-col + halfcols, row + halfrows, stack, zone, state);
					canvas.setCell(-row + halfcols, col + halfrows, stack, zone, state);
				}
			}
			break;
		case 7: /* Half Sym Random */
			// Diagonal (on face, on corner)
			vx = vy = (vx + vy) / 2;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = -vy; row < 0; ++row)
			for (col = row; col < vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols, row + halfrows, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols, row + halfrows, stack, zone, state);
				canvas.setCell(row + halfcols, col + halfrows, stack, zone, state);
				canvas.setCell(-col + halfcols, -row + halfrows, stack, zone, state);
				canvas.setCell(-row + halfcols, -col + halfrows, stack, zone, state);
			}
			break;
		case 8: /* Full Antisym Random */
			// Square
			vx = vy = (vx + vy) / 2;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack) {
				for (row = -vy; row < 0; ++row)
				for (col = row; col < vx; ++col)
				if (Life4DCanvas.cellCheck(polychoron, col + halfcols, row + halfrows, stack, zone)
						&& generator.nextInt(100) < percent) {
					canvas.setCell(col + halfcols + even, row + halfrows + even, stack, zone, state);
					canvas.setCell(row + halfcols + even, -col + halfrows, stack, zone, state);
					canvas.setCell(-col + halfcols, -row + halfrows, stack, zone, state);
					canvas.setCell(-row + halfcols, col + halfrows + even, stack, zone, state);
				}
				if (Life4DCanvas.cellCheck(polychoron, halfcols, halfrows, stack, zone)
						&& even == 0 && generator.nextInt(100) < percent) {
					canvas.setCell(halfcols, halfrows, stack, zone, state);
				}
			}
			break;
		case 9: /* Full Sym Random */
			// Square
			vy = (vx + vy) / 2;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack) {
				for (row = -vy; row < 0; ++row)
				for (col = row; col <= 0; ++col) {
				if (Life4DCanvas.cellCheck(polychoron, col + halfcols, row + halfrows, stack, zone)
						&& generator.nextInt(100) < percent) {
					canvas.setCell(col + halfcols + even, row + halfrows + even, stack, zone, state);
					canvas.setCell(row + halfcols + even, -col + halfrows, stack, zone, state);
					canvas.setCell(-col + halfcols, -row + halfrows, stack, zone, state);
					canvas.setCell(-row + halfcols, col + halfrows + even, stack, zone, state);
					canvas.setCell(-col + halfcols, row + halfrows + even, stack, zone, state);
					canvas.setCell(-row + halfcols, -col + halfrows, stack, zone, state);
					canvas.setCell(col + halfcols + even, -row + halfrows, stack, zone, state);
					canvas.setCell(row + halfcols + even, col + halfrows + even, stack, zone, state);
				}
				if (Life4DCanvas.cellCheck(polychoron, halfcols, halfrows, stack, zone)
						&& even == 0 && generator.nextInt(100) < percent) {
					canvas.setCell(halfcols, halfrows, stack, zone, state);
				}
				}
			}
			break;
		case 10: /* Half Antisym Skew Random */
			// Skew Diagonal
			vx = vy = vz = (vx + vy + vz) / 3;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = -vz; stack < vz; ++stack)
			for (row = stack; row < vy; ++row)
			for (col = row; col < vx; ++col)
			if (even == 1) {
				if (Life4DCanvas.cellCheck(polychoron, col + halfcols, row + halfrows, stack + halfstacks, zone)
						&& generator.nextInt(100) < percent) {
					canvas.setCell(col + halfcols, row + halfrows, stack + halfstacks, zone, state);
					canvas.setCell(row + halfcols, stack + halfrows, col + halfstacks, zone, state);
					canvas.setCell(stack + halfcols, col + halfrows, row + halfstacks, zone, state);
				}
			} else {
				if (Life4DCanvas.cellCheck(polychoron, -col + halfcols, row + halfrows, stack + halfstacks, zone)
						&& generator.nextInt(100) < percent) {
					canvas.setCell(-col + halfcols, row + halfrows, stack + halfstacks, zone, state);
					canvas.setCell(-row + halfcols, stack + halfrows, col + halfstacks, zone, state);
						canvas.setCell(-stack + halfcols, col + halfrows, row + halfstacks, zone, state);
				}
			}
			break;
		case 11: /* Sym XYZ Random */
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks; stack < halfstacks + vz; ++stack)
			for (row = halfrows; row < halfrows + vy; ++row)
			for (col = halfcols; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, 2 * halfrows - row - yEven, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
				canvas.setCell(col, 2 * halfrows - row - yEven, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, row, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, stack, zone, state);
				canvas.setCell(col, row, 2 * halfstacks - stack - zEven, zone, state);
				canvas.setCell(col, 2 * halfrows - row - yEven, 2 * halfstacks - stack - zEven, zone, state);
				canvas.setCell(2 * halfcols - col - even, row, 2 * halfstacks - stack - zEven, zone, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, 2 * halfstacks - stack - zEven, zone, state);
			}
			break;
		case 12: /* Half Sym Skew Random */
			// Skew Diagonal (on face, on corner)
			vx = vy = vz = (vx + vy + vz) / 3;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = -vz; stack < 0; ++stack)
			for (row = stack; row < vy; ++row)
			for (col = row; col < vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols, row + halfrows, stack + halfstacks, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols, row + halfrows, stack + halfstacks, zone, state);
				canvas.setCell(row + halfcols, stack + halfrows, col + halfstacks, zone, state);
				canvas.setCell(stack + halfcols, col + halfrows, row + halfstacks, zone, state);
				canvas.setCell(-col + halfcols, -row + halfrows, -stack + halfstacks, zone, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, -col + halfstacks, zone, state);
				canvas.setCell(-stack + halfcols, -col + halfrows, -row + halfstacks, zone, state);
			}
			break;
		case 13: /* Full Antisym Cube Random */
			vx = vy = vz = (vx + vy + vz) / 3;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = -vz; stack < 0; ++stack)
			for (row = stack; row < vy; ++row)
			for (col = row; col < vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, col + halfstacks + even, zone, state);
				canvas.setCell(stack + halfcols + even, col + halfrows + even, row + halfstacks + even, zone, state);
				canvas.setCell(-col + halfcols, -row + halfrows, -stack + halfstacks, zone, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, -col + halfstacks, zone, state);
				canvas.setCell(-stack + halfcols, -col + halfrows, -row + halfstacks, zone, state);
			}
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			if (Life4DCanvas.cellCheck(polychoron, halfcols, halfrows, halfstacks, zone)
					&& even == 0 && generator.nextInt(100) < percent) {
				canvas.setCell(halfcols, halfrows, halfstacks, zone, state);
			}
			break;
		case 14: /* Full Sym Cube Random */
			vz = (vx + vy + vz) / 3;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = -vz; stack < 0; ++stack)
			for (row = stack; row <= 0; ++row)
			for (col = row; col <= 0; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, col + halfstacks + even, zone, state);
				canvas.setCell(stack + halfcols + even, col + halfrows + even, row + halfstacks + even, zone, state);
				canvas.setCell(-col + halfcols, -row + halfrows, -stack + halfstacks, zone, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, -col + halfstacks, zone, state);
				canvas.setCell(-stack + halfcols, -col + halfrows, -row + halfstacks, zone, state);
				canvas.setCell(-col + halfcols, row + halfrows + even, stack + halfstacks + even, zone, state);
				canvas.setCell(-row + halfcols, stack + halfrows + even, col + halfstacks + even, zone, state);
				canvas.setCell(-stack + halfcols, col + halfrows + even, row + halfstacks + even, zone, state);
				canvas.setCell(col + halfcols + even, -row + halfrows, -stack + halfstacks, zone, state);
				canvas.setCell(row + halfcols + even, -stack + halfrows, -col + halfstacks, zone, state);
				canvas.setCell(stack + halfcols + even, -col + halfrows, -row + halfstacks, zone, state);
				canvas.setCell(col + halfcols + even, -row + halfrows, stack + halfstacks + even, zone, state);
				canvas.setCell(row + halfcols + even, -stack + halfrows, col + halfstacks + even, zone, state);
				canvas.setCell(stack + halfcols + even, -col + halfrows, row + halfstacks + even, zone, state);
				canvas.setCell(-col + halfcols, row + halfrows + even, -stack + halfstacks, zone, state);
				canvas.setCell(-row + halfcols, stack + halfrows + even, -col + halfstacks, zone, state);
				canvas.setCell(-stack + halfcols, col + halfrows + even, -row + halfstacks, zone, state);
				canvas.setCell(col + halfcols + even, row + halfrows + even, -stack + halfstacks, zone, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, -col + halfstacks, zone, state);
				canvas.setCell(stack + halfcols + even, col + halfrows + even, -row + halfstacks, zone, state);
				canvas.setCell(-col + halfcols, -row + halfrows, stack + halfstacks + even, zone, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, col + halfstacks + even, zone, state);
				canvas.setCell(-stack + halfcols, -col + halfrows, row + halfstacks + even, zone, state);
			}
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
				if (Life4DCanvas.cellCheck(polychoron, halfcols, halfrows, halfstacks, zone)
						&& even == 0 && generator.nextInt(100) < percent) {
					canvas.setCell(halfcols, halfrows, halfstacks, zone, state);
			}
			break;
		case 15: /* Full Tetra Random */
			vz = (vx + vy + vz) / 3;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = -vz; stack < 0; ++stack)
			for (row = stack; row <= 0; ++row)
			for (col = row; col <= 0; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, col + halfstacks + even, zone, state);
				canvas.setCell(stack + halfcols + even, col + halfrows + even, row + halfstacks + even, zone, state);
				canvas.setCell(col + halfcols + even, -row + halfrows, -stack + halfstacks, zone, state);
				canvas.setCell(row + halfcols + even, -stack + halfrows, -col + halfstacks, zone, state);
				canvas.setCell(stack + halfcols + even, -col + halfrows, -row + halfstacks, zone, state);
				canvas.setCell(-col + halfcols, row + halfrows + even, -stack + halfstacks, zone, state);
				canvas.setCell(-row + halfcols, stack + halfrows + even, -col + halfstacks, zone, state);
				canvas.setCell(-stack + halfcols, col + halfrows + even, -row + halfstacks, zone, state);
				canvas.setCell(-col + halfcols, -row + halfrows, stack + halfstacks + even, zone, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, col + halfstacks + even, zone, state);
				canvas.setCell(-stack + halfcols, -col + halfrows, row + halfstacks + even, zone, state);
			}
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			if (Life4DCanvas.cellCheck(polychoron, halfcols, halfrows, halfstacks, zone)
					&& even == 0 && generator.nextInt(100) < percent) {
				canvas.setCell(halfcols, halfrows, halfstacks, zone, state);
			}
			break;
		case 16: /* Alt Full Antisym Cube Random */
			vx = vy = vz = (vx + vy + vz) / 3;
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = -vz; stack < 0; ++stack)
			for (row = stack; row < vy; ++row)
			for (col = row; col < vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, col + halfstacks + even, zone, state);
				canvas.setCell(stack + halfcols + even, col + halfrows + even, row + halfstacks + even, zone, state);
				canvas.setCell(-col + halfcols, -stack + halfrows, -row + halfstacks, zone, state);
				canvas.setCell(-row + halfcols, -col + halfrows, -stack + halfstacks, zone, state);
				canvas.setCell(-stack + halfcols, -row + halfrows, -col + halfstacks, zone, state);
			}
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			if (Life4DCanvas.cellCheck(polychoron, halfcols, halfrows, halfstacks, zone)
					&& even == 0 && generator.nextInt(100) < percent) {
				canvas.setCell(halfcols, halfrows, halfstacks, zone, state);
			}
			break;
		case 17: /* Antisym Z Random */
			if (polychoron == 24) {
				even = yEven = generator.nextInt(2);
			}
			for (zone = -vw; zone < vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = halfrows; row < halfrows + vy; ++row)
			for (col = halfcols - vx; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone + halfzones)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone + halfzones, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, 2 * halfstacks - stack, zone + halfzones, state);
			}
			break;
		case 18: /* Half Antisym Metaskew Random */
			// Skew Diagonal
			vx = vy = vz = vw = (vx + vy + vz + vw) / 4;
			for (zone = -vw; zone < vw; ++zone)
			for (stack = zone; stack < vz; ++stack)
			for (row = stack; row < vy; ++row)
			for (col = row; col < vx; ++col)
			if (even == 1) {
					if (generator.nextInt(100) < percent
							&& Life4DCanvas.cellCheck(polychoron, col + halfcols, row + halfrows, stack + halfstacks, zone + halfzones)) {
						canvas.setCell(col + halfcols, row + halfrows, stack + halfstacks, zone + halfzones, state);
						canvas.setCell(row + halfcols, stack + halfrows, zone + halfstacks, col + halfzones, state);
						canvas.setCell(stack + halfcols, zone + halfrows, col + halfstacks, row + halfzones, state);
						canvas.setCell(zone + halfcols, col + halfrows, row + halfstacks, stack + halfzones, state);
					}
			} else {
					if (generator.nextInt(100) < percent
							&& Life4DCanvas.cellCheck(polychoron, -col + halfcols, row + halfrows, stack + halfstacks, zone + halfzones)) {
						canvas.setCell(-col + halfcols, row + halfrows, stack + halfstacks, zone + halfzones, state);
						canvas.setCell(-row + halfcols, stack + halfrows, zone + halfstacks, col + halfzones, state);
						canvas.setCell(-stack + halfcols, zone + halfrows, col + halfstacks, row + halfzones, state);
						canvas.setCell(-zone + halfcols, col + halfrows, row + halfstacks, stack + halfzones, state);
					}
			}
			break;
		case 19: /* Sym XYZW Random */
			for (zone = halfzones; zone < halfzones + vw; ++zone)
			for (stack = halfstacks; stack < halfstacks + vz; ++stack)
			for (row = halfrows; row < halfrows + vy; ++row)
			for (col = halfcols; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
				canvas.setCell(col, 2 * halfrows - row - yEven, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, row, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, stack, zone, state);
				canvas.setCell(col, row, 2 * halfstacks - stack - zEven, zone, state);
				canvas.setCell(col, 2 * halfrows - row - yEven, 2 * halfstacks - stack - zEven, zone, state);
				canvas.setCell(2 * halfcols - col - even, row, 2 * halfstacks - stack - zEven, zone, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, 2 * halfstacks - stack - zEven, zone, state);
				canvas.setCell(col, row, stack, 2 * halfzones - zone - wEven, state);
				canvas.setCell(col, 2 * halfrows - row - yEven, stack, 2 * halfzones - zone - wEven, state);
				canvas.setCell(2 * halfcols - col - even, row, stack, 2 * halfzones - zone - wEven, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, stack, 2 * halfzones - zone - wEven, state);
				canvas.setCell(col, row, 2 * halfstacks - stack - zEven, 2 * halfzones - zone - wEven, state);
				canvas.setCell(col, 2 * halfrows - row - yEven, 2 * halfstacks - stack - zEven, 2 * halfzones - zone - wEven, state);
				canvas.setCell(2 * halfcols - col - even, row, 2 * halfstacks - stack - zEven, 2 * halfzones - zone - wEven, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, 2 * halfstacks - stack - zEven, 2 * halfzones - zone - wEven, state);
			}
			break;
		case 20: /* Half Sym Metaskew Random */
			// Skew Diagonal (on face, on corner)
			vx = vy = vz = vw = (vx + vy + vz + vw) / 4;
			for (zone = -vw; zone < 0; ++zone)
			for (stack = zone; stack < vz; ++stack)
			for (row = stack; row < vy; ++row)
			for (col = row; col < vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols, row + halfrows, stack + halfstacks, zone + halfzones, state);
				canvas.setCell(row + halfcols, stack + halfrows, zone + halfstacks, col + halfzones, state);
				canvas.setCell(stack + halfcols, zone + halfrows, col + halfstacks, row + halfzones, state);
				canvas.setCell(zone + halfcols, col + halfrows, row + halfstacks, stack + halfzones, state);
				canvas.setCell(-col + halfcols, -row + halfrows, -stack + halfstacks, -zone + halfzones, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, -zone + halfstacks, -col + halfzones, state);
				canvas.setCell(-stack + halfcols, -zone + halfrows, -col + halfstacks, -row + halfzones, state);
				canvas.setCell(-zone + halfcols, -col + halfrows, -row + halfstacks, -stack + halfzones, state);
			}
			break;
		case 21: /* Full Antisym Tesseract Random */
			vx = vy = vz = vw = (vx + vy + vz + vw) / 4;
			for (zone = -vw; zone < 0; ++zone)
			for (stack = zone; stack < vz; ++stack)
			for (row = stack; row < vy; ++row)
			for (col = row; col < vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone + halfzones + even)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone + halfzones + even, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, zone + halfstacks + even, col + halfzones + even, state);
				canvas.setCell(stack + halfcols + even, zone + halfrows + even, col + halfstacks + even, row + halfzones + even, state);
				canvas.setCell(zone + halfcols + even, col + halfrows + even, row + halfstacks + even, stack + halfzones + even, state);
				canvas.setCell(-col + halfcols, -row + halfrows, -stack + halfstacks, -zone + halfzones, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, -zone + halfstacks, -col + halfzones, state);
				canvas.setCell(-stack + halfcols, -zone + halfrows, -col + halfstacks, -row + halfzones, state);
			}
			if (even == 0 && generator.nextInt(100) < percent) {
				canvas.setCell(halfcols, halfrows, halfstacks, halfzones, state);
			}
			break;
		case 22: /* Full Sym Tesseract Random */
			vw =  (vx + vy + vz + vw) / 4;
			for (zone = -vw; zone < 0; ++zone)
			for (stack = zone; stack <= 0; ++stack)
			for (row = stack; row <= 0; ++row)
			for (col = row; col <= 0; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone + halfzones + even)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone + halfzones + even, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, zone + halfstacks + even, col + halfzones + even, state);
				canvas.setCell(stack + halfcols + even, zone + halfrows + even, col + halfstacks + even, row + halfzones + even, state);
				canvas.setCell(zone + halfcols + even, col + halfrows + even, row + halfstacks + even, stack + halfzones + even, state);
				canvas.setCell(-col + halfcols, -row + halfrows, -stack + halfstacks, -zone + halfzones, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, -zone + halfstacks, -col + halfzones, state);
				canvas.setCell(-stack + halfcols, -zone + halfrows, -col + halfstacks, -row + halfzones, state);
				canvas.setCell(-zone + halfcols, -col + halfrows, -row + halfstacks, -stack + halfzones, state);
				canvas.setCell(-col + halfcols, row + halfrows + even, stack + halfstacks + even, zone + halfzones + even, state);
				canvas.setCell(-row + halfcols, stack + halfrows + even, zone + halfstacks + even, col + halfzones + even, state);
				canvas.setCell(-stack + halfcols, zone + halfrows + even, col + halfstacks + even, row + halfzones + even, state);
				canvas.setCell(-zone + halfcols, col + halfrows + even, row + halfstacks + even, stack + halfzones + even, state);
				canvas.setCell(col + halfcols + even, -row + halfrows, -stack + halfstacks, -zone + halfzones, state);
				canvas.setCell(row + halfcols + even, -stack + halfrows, -zone + halfstacks, -col + halfzones, state);
				canvas.setCell(stack + halfcols + even, -zone + halfrows, -col + halfstacks, -row + halfzones, state);
				canvas.setCell(zone + halfcols + even, -col + halfrows, -row + halfstacks, -stack + halfzones, state);
				canvas.setCell(col + halfcols + even, -row + halfrows, stack + halfstacks + even, zone + halfzones + even, state);
				canvas.setCell(row + halfcols + even, -stack + halfrows, zone + halfstacks + even, col + halfzones + even, state);
				canvas.setCell(stack + halfcols + even, -zone + halfrows, col + halfstacks + even, row + halfzones + even, state);
				canvas.setCell(zone + halfcols + even, -col + halfrows, row + halfstacks + even, stack + halfzones + even, state);
				canvas.setCell(-col + halfcols, row + halfrows + even, -stack + halfstacks, -zone + halfzones, state);
				canvas.setCell(-row + halfcols, stack + halfrows + even, -zone + halfstacks, -col + halfzones, state);
				canvas.setCell(-stack + halfcols, zone + halfrows + even, -col + halfstacks, -row + halfzones, state);
				canvas.setCell(-zone + halfcols, col + halfrows + even, -row + halfstacks, -stack + halfzones, state);
				canvas.setCell(col + halfcols + even, row + halfrows + even, -stack + halfstacks, zone + halfzones + even, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, -zone + halfstacks, col + halfzones + even, state);
				canvas.setCell(stack + halfcols + even, zone + halfrows + even, -col + halfstacks, row + halfzones + even, state);
				canvas.setCell(zone + halfcols + even, col + halfrows + even, -row + halfstacks, stack + halfzones + even, state);
				canvas.setCell(-col + halfcols, -row + halfrows, stack + halfstacks + even, -zone + halfzones, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, zone + halfstacks + even, -col + halfzones, state);
				canvas.setCell(-stack + halfcols, -zone + halfrows, col + halfstacks + even, -row + halfzones, state);
				canvas.setCell(-zone + halfcols, -col + halfrows, row + halfstacks + even, -stack + halfzones, state);
				canvas.setCell(-col + halfcols, row + halfrows + even, -stack + halfstacks, zone + halfzones + even, state);
				canvas.setCell(-row + halfcols, stack + halfrows + even, -zone + halfstacks, col + halfzones + even, state);
				canvas.setCell(-stack + halfcols, zone + halfrows + even, -col + halfstacks, row + halfzones + even, state);
				canvas.setCell(-zone + halfcols, col + halfrows + even, -row + halfstacks, stack + halfzones + even, state);
				canvas.setCell(col + halfcols + even, -row + halfrows, stack + halfstacks + even, -zone + halfzones, state);
				canvas.setCell(row + halfcols + even, -stack + halfrows, zone + halfstacks + even, -col + halfzones, state);
				canvas.setCell(stack + halfcols + even, -zone + halfrows, col + halfstacks + even, -row + halfzones, state);
				canvas.setCell(zone + halfcols + even, -col + halfrows, row + halfstacks + even, -stack + halfzones, state);
				canvas.setCell(-col + halfcols, -row + halfrows, stack + halfstacks + even, zone + halfzones + even, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, zone + halfstacks + even, col + halfzones + even, state);
				canvas.setCell(-stack + halfcols, -zone + halfrows, col + halfstacks + even, row + halfzones + even, state);
				canvas.setCell(-zone + halfcols, -col + halfrows, row + halfstacks + even, stack + halfzones + even, state);
				canvas.setCell(col + halfcols + even, row + halfrows + even, -stack + halfstacks, -zone + halfzones, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, -zone + halfstacks, -col + halfzones, state);
				canvas.setCell(stack + halfcols + even, zone + halfrows + even, -col + halfstacks, -row + halfzones, state);
				canvas.setCell(zone + halfcols + even, col + halfrows + even, -row + halfstacks, -stack + halfzones, state);
			}
			if (Life4DCanvas.cellCheck(polychoron, halfcols, halfrows, halfstacks, halfzones)
					&& even == 0 && generator.nextInt(100) < percent) {
				canvas.setCell(halfcols, halfrows, halfstacks, halfzones, state);
			}
			break;
		case 23: /* Full Penta Random */
			vw = (vx + vy + vz + vw) / 4;
			for (zone = -vw; zone < 0; ++zone)
			for (stack = zone; stack <= 0; ++stack)
			for (row = stack; row <= 0; ++row)
			for (col = row; col <= 0; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone + halfzones + even)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col + halfcols + even, row + halfrows + even, stack + halfstacks + even, zone + halfzones + even, state);
				canvas.setCell(row + halfcols + even, stack + halfrows + even, zone + halfstacks + even, col + halfzones + even, state);
				canvas.setCell(stack + halfcols + even, zone + halfrows + even, col + halfstacks + even, row + halfzones + even, state);
				canvas.setCell(zone + halfcols + even, col + halfrows + even, row + halfstacks + even, stack + halfzones + even, state);
				canvas.setCell(col + halfcols + even, -row + halfrows, -stack + halfstacks, -zone + halfzones, state);
				canvas.setCell(row + halfcols + even, -stack + halfrows, -zone + halfstacks, -col + halfzones, state);
				canvas.setCell(stack + halfcols + even, -zone + halfrows, -col + halfstacks, -row + halfzones, state);
				canvas.setCell(zone + halfcols + even, -col + halfrows, -row + halfstacks, -stack + halfzones, state);
				canvas.setCell(-col + halfcols, row + halfrows + even, -stack + halfstacks, -zone + halfzones, state);
				canvas.setCell(-row + halfcols, stack + halfrows + even, -zone + halfstacks, -col + halfzones, state);
				canvas.setCell(-stack + halfcols, zone + halfrows + even, -col + halfstacks, -row + halfzones, state);
				canvas.setCell(-zone + halfcols, col + halfrows + even, -row + halfstacks, -stack + halfzones, state);
				canvas.setCell(-col + halfcols, -row + halfrows, stack + halfstacks + even, -zone + halfzones, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, zone + halfstacks + even, -col + halfzones, state);
				canvas.setCell(-stack + halfcols, -zone + halfrows, col + halfstacks + even, -row + halfzones, state);
				canvas.setCell(-zone + halfcols, -col + halfrows, row + halfstacks + even, -stack + halfzones, state);
				canvas.setCell(-col + halfcols, -row + halfrows, -stack + halfstacks, zone + halfzones + even, state);
				canvas.setCell(-row + halfcols, -stack + halfrows, -zone + halfstacks, col + halfzones + even, state);
				canvas.setCell(-stack + halfcols, -zone + halfrows, -col + halfstacks, row + halfzones + even, state);
				canvas.setCell(-zone + halfcols, -col + halfrows, -row + halfstacks, stack + halfzones + even, state);
			}
			if (Life4DCanvas.cellCheck(polychoron, halfcols, halfrows, halfstacks, halfzones)
					&& even == 0 && generator.nextInt(100) < percent) {
				canvas.setCell(halfcols, halfrows, halfstacks, halfzones, state);
			}
			break;
		case 24: /* Antisym W Random */
			if (polychoron == 24) {
				even = yEven = generator.nextInt(2);
			}
			for (zone = halfzones - vw; zone < halfzones + vw; ++zone)
			for (stack = halfstacks - vz; stack < halfstacks + vz; ++stack)
			for (row = halfrows; row < halfrows + vy; ++row)
			for (col = halfcols - vx; col < halfcols + vx; ++col)
			if (Life4DCanvas.cellCheck(polychoron, col, row, stack, zone)
					&& generator.nextInt(100) < percent) {
				canvas.setCell(col, row, stack, zone, state);
				canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, 2 * halfstacks - stack, 2 * halfzones - zone, state);
			}
			break;
		default:
			System.out.println("wrong symmetry index " +
				symmetryIndex);
		}
	}
}
