package com.sillycycle.bagleyd.util;

/*
 * @(#)Box4D.java
 * Copied from
 * @(#)Rectangle.java 1.70 04/05/18
 * through some code away like intersections, etc.
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

//package java.awt;

import java.awt.Rectangle;

/**
* A <code>Box</code> specifies an area in a coordinate space that is
* enclosed by the <code>Box</code> object's top-left point
* (<i>x</i>,&nbsp;<i>y</i>)
* in the coordinate space, its width, its height, its depth, and its zone.
* <p>
* A <code>Box</code> object's <code>width</code> and
* <code>height</code> are <code>public</code> fields. The constructors
* that create a <code>Box</code>, and the methods that can modify
* one, do not prevent setting a negative value for width or height.
* <p>
* A <code>Box</code> whose width or height is negative is considered
* empty. If the <code>Box</code> is empty, then the
* <code>isEmpty</code> method returns <code>true</code>. No point can be
* contained by or inside an empty <code>Box</code>. The
* values of <code>width</code> and <code>height</code>, however, are still
* valid. An empty <code>Box</code> still has a location in the
* coordinate space, and methods that change its size or location remain
* valid. The behavior of methods that operate on more than one
* <code>Box</code> is undefined if any of the participating
* <code>Box</code> objects has a negative
* <code>width</code> or <code>height</code>. These methods include
* <code>intersects</code>, <code>intersection</code>, and
* <code>union</code>.
*
* @version 1.70, 05/18/04
* @author Sami Shaio
* @since JDK1.0
*/
public class Box4D extends Rectangle {

	/**
	 * The <i>x</i> coordinate of the <code>Box</code>.
	 *
	 * @serial
	 * @see #setLocation(int, int)
	 * @see #getLocation()
	 */
	//public int x;

	/**
	 * The <i>y</i> coordinate of the <code>Box</code>.
	 *
	 * @serial
	 * @see #setLocation(int, int)
	 * @see #getLocation()
	 */
	//public int y;

	/**
	 * The <i>z</i> coordinate of the <code>Box</code>.
	 *
	 * @serial
	 * @see #setLocation(int, int)
	 * @see #getLocation()
	 */
	public int z;

	/**
	 * The <i>z</i> coordinate of the <code>Box</code>.
	 *
	 * @serial
	 * @see #setLocation(int, int)
	 * @see #getLocation()
	 */
	public int w;

	/**
	 * The width of the <code>Box</code>.
	 * @serial
	 * @see #setSize(int, int)
	 * @see #getSize()
	 * @since JDK1.0.
	 */
	//public int width;

	/**
	 * The height of the <code>Box</code>.
	 *
	 * @serial
	 * @see #setSize(int, int)
	 * @see #getSize()
	 */
	//public int height;

	/**
	 * The depth of the <code>Box</code>.
	 *
	 * @serial
	 * @see #setSize(int, int)
	 * @see #getSize()
	 */
	public int depth;

	/**
	 * The zone of the <code>Box</code>.
	 *
	 * @serial
	 * @see #setSize(int, int)
	 * @see #getSize()
	 */
	public int zone;

	/*
	 * JDK 1.1 serialVersionUID
	 */
	 private static final long serialVersionUID = -4345857070255674764L;

	/**
	 * Constructs a new <code>Box</code> whose front-top-left corner
	 * is at (0,&nbsp;0,&nbsp;0) in the coordinate space, and whose width,
	 * height, depth, and zone are both zero.
	 */
	public Box4D() {
		this(0, 0, 0, 0, 0, 0, 0, 0);
	}

	/**
	 * Constructs a new <code>Box</code>, initialized to match
	 * the values of the specified <code>Box</code>.
	 * @param r the <code>Box</code> from which to copy initial values
	 * to a newly constructed <code>Box</code>
	 * @since JDK1.1
	 */
	public Box4D(Box4D  r) {
		this(r.x, r.y, r.z, r.w, r.width, r.height, r.depth, r.zone);
	}

	/**
	  * Constructs a new <code>Box</code> whose front-top-left corner is
	  * specified as
	  * (<code>x</code>,&nbsp;<code>y</code>) and whose width and height
	  * are specified by the arguments of the same name.
	  * @param x the specified x coordinate
	  * @param y the specified y coordinate
	  * @param z the specified z coordinate
	  * @param w the specified w coordinate
	  * @param width the width of the <code>Box</code>
	  * @param height the height of the <code>Box</code>
	  * @param depth the depth of the <code>Box</code>
	  * @param zone the zone of the <code>Box</code>
	  */
	public Box4D(int x, int y, int z, int w, int width, int height, int depth, int zone) {
		super.x = x;
		super.y = y;
		this.z = z;
		this.w = w;
		super.width = width;
		super.height = height;
		this.depth = depth;
		this.zone = zone;
	}

	/**
	 * Constructs a new <code>Box</code> whose front-top-left corner
	 * is at (0,&nbsp;0,&nbsp;0) in the coordinate space, and whose width,
	 * height, and depth are specified by the arguments of the same name.
	 * @param width the width of the <code>Box</code>
	 * @param height the height of the <code>Box</code>
	 * @param depth the depth of the <code>Box</code>
	 * @param zone the zone of the <code>Box</code>
	 */
	public Box4D(int width, int height, int depth, int zone) {
		this(0, 0, 0, 0, width, height, depth, zone);
	}

	/**
	 * Constructs a new <code>Box</code> whose front-top-left corner is
	 * specified by the {@link Point} argument, and
	 * whose width, height, and depth are specified by the
	 * {@link Dimension} argument.
	 * @param p a <code>Point</code> that is the front-top-left corner of
	 * the <code>Box</code>
	 * @param d a <code>Dimension</code>, representing the
	 * width, height, depth, and zone of the <code>Box</code>
	 */
	public Box4D(Point4D p, Dimension4D d) {
		this(p.x, p.y, p.z, p.w, d.width, d.height, d.depth, d.zone);
	}

	/**
	 * Constructs a new <code>Box</code> whose front-top-left corner is the
	 * specified <code>Point</code>, and whose width, height, and depth
	 * are all zero.
	 * @param p a <code>Point</code> that is the top left corner
	 * of the <code>Box</code>
	 */
	public Box4D(Point4D p) {
		this(p.x, p.y, p.z, p.w, 0, 0, 0, 0);
	}

	/**
	 * Constructs a new <code>Box</code> whose top left corner is
	 * (0,&nbsp;0) and whose width and height are specified
	 * by the <code>Dimension</code> argument.
	 * @param d a <code>Dimension</code>, specifying width, height,
	 * and depth
	 */
	public Box4D(Dimension4D d) {
		this(0, 0, 0, 0, d.width, d.height, d.depth, d.zone);
	}

	/**
	 * Returns the X coordinate of the bounding <code>Box</code> in
	 * <code>double</code> precision.
	 * @return the x coordinate of the bounding <code>Box</code>.
	 */
	@Override
	public double getX() {
		return x;
	}

	/**
	 * Returns the Y coordinate of the bounding <code>Box</code> in
	 * <code>double</code> precision.
	 * @return the y coordinate of the bounding <code>Box</code>.
	 */
	@Override
	public double getY() {
		return y;
	}

	/**
	 * Returns the Z coordinate of the bounding <code>Box</code> in
	 * <code>double</code> precision.
	 * @return the z coordinate of the bounding <code>Box</code>.
	 */
	public double getZ() {
		return z;
	}

	/**
	 * Returns the W coordinate of the bounding <code>Box</code> in
	 * <code>double</code> precision.
	 * @return the w coordinate of the bounding <code>Box</code>.
	 */
	public double getW() {
		return w;
	}

	/**
	 * Returns the width of the bounding <code>Box</code> in
	 * <code>double</code> precision.
	 * @return the width of the bounding <code>Box</code>.
	 */
	@Override
	public double getWidth() {
		return width;
	}

	/**
	 * Returns the height of the bounding <code>Box</code> in
	 * <code>double</code> precision.
	 * @return the height of the bounding <code>Box</code>.
	 */
	@Override
	public double getHeight() {
		return height;
	}

	/**
	 * Returns the depth of the bounding <code>Box</code> in
	 * <code>double</code> precision.
	 * @return the depth of the bounding <code>Box</code>.
	 */
	public double getDepth() {
		return depth;
	}

	/**
	 * Returns the zone of the bounding <code>Box</code> in
	 * <code>double</code> precision.
	 * @return the zone of the bounding <code>Box</code>.
	 */
	public double getZone() {
		return zone;
	}

	/**
	 * Gets the bounding <code>Box</code> of this <code>Box</code>.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>getBounds</code> method of
	 * {@link Component}.
	 * @return a new <code>Box</code>, equal to the
	 * bounding <code>Box</code> for this <code>Box</code>.
	 * @see java.awt.Component#getBounds
	 * @see #setBounds(Box)
	 * @see #setBounds(int, int, int, int, int, int, int, int)
	 * @since JDK1.1
	 */
	@Override
	public Box4D getBounds() {
		return new Box4D (x, y, z, w, width, height, depth, zone);
	}

	/**
	 * Return the high precision bounding box of this rectangle.
	 * @since 1.2
	 */
	@Override
	public Rectangle getBounds2D() {
		return new Rectangle (x, y, width, height);
	}

	/**
	 * Sets the bounding <code>Box</code> of this <code>Box</code>
	 * to match the specified <code>Box</code>.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>setBounds</code> method of <code>Component</code>.
	 * @param r the specified <code>Box</code>
	 * @see #getBounds
	 * @see java.awt.Component#setBounds(java.awt.Box)
	 * @since JDK1.1
	 */
	public void setBounds(Box4D b) {
		setBounds(b.x, b.y, b.z, b.w, b.width, b.height, b.depth, b.zone);
	}

	/**
	 * Sets the bounding <code>Box</code> of this
	 * <code>Box</code> to the specified
	 * <code>x</code>, <code>y</code>, <code>width</code>,
	 * and <code>height</code>.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>setBounds</code> method of <code>Component</code>.
	 * @param x the new x coordinate for the front-top-left
	 * corner of this <code>Box</code>
	 * @param y the new y coordinate for the front-top-left
	 * corner of this <code>Box</code>
	 * @param z the new z coordinate for the front-top-left
	 * corner of this <code>Box</code>
	 * @param w the new w coordinate for the front-top-left
	 * corner of this <code>Box</code>
	 * @param width the new width for this <code>Box</code>
	 * @param height the new height for this <code>Box</code>
	 * @param depth the new depth for this <code>Box</code>
	 * @param zone the new zone for this <code>Box</code>
	 * @see #getBounds
	 * @see java.awt.Component#setBounds(int, int, int, int, int, int, int, int)
	 * @since JDK1.1
	 */
	public void setBounds(int x, int y, int z, int w, int width, int height, int depth, int zone) {
		reshape(x, y, z, w, width, height, depth, zone);
	}

	/**
	 * Sets the bounds of this <code>Box</code> to the specified
	 * <code>x</code>, <code>y</code>, <code>width</code>,
	 * and <code>height</code>.
	 * This method is included for completeness, to parallel the
	 * <code>setBounds</code> method of <code>Component</code>.
	 * @param x the x coordinate of the upper-left corner of
	 * the specified rectangle
	 * @param y the y coordinate of the upper-left corner of
	 * the specified rectangle
	 * @param width the new width for the <code>Dimension</code> object
	 * @param height the new height for the <code>Dimension</code> object
	 */
	public void setRect(double x, double y, double z, double w, double width, double height, double depth, double zone) {
		int x0 = (int) Math.floor(x);
		int y0 = (int) Math.floor(y);
		int z0 = (int) Math.floor(z);
		int w0 = (int) Math.floor(w);
		int x1 = (int) Math.ceil(x+width);
		int y1 = (int) Math.ceil(y+height);
		int z1 = (int) Math.ceil(z+depth);
		int w1 = (int) Math.ceil(w+zone);
		setBounds(x0, y0, z0, w0, x1-x0, y1-y0, z1-z0, w1-w0);
	}

	/**
	 * Sets the bounding <code>Box</code> of this
	 * <code>Box</code> to the specified
	 * <code>x</code>, <code>y</code>, <code>width</code>,
	 * and <code>height</code>.
	 * <p>
	 * @param x the new x coordinate for the front-top-left
	 * corner of this <code>Box</code>
	 * @param y the new y coordinate for the front-top-left
	 * corner of this <code>Box</code>
	 * @param z the new z coordinate for the front-top-left
	 * corner of this <code>Box</code>
	 * @param w the new w coordinate for the front-top-left
	 * corner of this <code>Box</code>
	 * @param width the new width for this <code>Box</code>
	 * @param height the new height for this <code>Box</code>
	 * @param depth the new depth for this <code>Box</code>
	 * @param zone the new zone for this <code>Box</code>
	 * @deprecated As of JDK version 1.1,
	 * replaced by <code>setBounds(int, int, int, int)</code>.
	 */
	@Deprecated
	public void reshape(int x, int y, int z, int w, int width, int height, int depth, int zone) {
		super.x = x;
		super.y = y;
		this.z = z;
		this.w = w;
		super.width = width;
		super.height = height;
		this.depth = depth;
		this.zone = zone;
	}

	/**
	 * Returns the location of this <code>Box</code>.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>getLocation</code> method of <code>Component</code>.
	 * @return the <code>Point</code> that is the front-top-left corner of
	 * this <code>Box</code>.
	 * @see java.awt.Component#getLocation
	 * @see #setLocation(Point)
	 * @see #setLocation(int, int)
	 * @since JDK1.1
	 */
	@Override
	public Point4D getLocation() {
		return new Point4D(x, y, z, w);
	}

	/**
	 * Moves this <code>Box</code> to the specified location.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>setLocation</code> method of <code>Component</code>.
	 * @param p the <code>Point</code> specifying the new location
	 * for this <code>Box</code>
	 * @see java.awt.Component#setLocation(java.awt.Point)
	 * @see #getLocation
	 * @since JDK1.1
	 */
	public void setLocation(Point4D p) {
		setLocation(p.x, p.y, p.z, p.w);
	}

	/**
	 * Moves this <code>Box</code> to the specified location.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>setLocation</code> method of <code>Component</code>.
	 * @param x the x coordinate of the new location
	 * @param y the y coordinate of the new location
	 * @param z the z coordinate of the new location
	 * @param w the w coordinate of the new location
	 * @see #getLocation
	 * @see java.awt.Component#setLocation(int, int, int, int)
	 * @since JDK1.1
	 */
	public void setLocation(int x, int y, int z, int w) {
		move(x, y, z, w);
	}

	/**
	 * Moves this <code>Box</code> to the specified location.
	 * <p>
	 * @param x the x coordinate of the new location
	 * @param y the y coordinate of the new location
	 * @param z the z coordinate of the new location
	 * @param w the w coordinate of the new location
	 * @deprecated As of JDK version 1.1,
	 * replaced by <code>setLocation(int, int, int, int)</code>.
	 */
	@Deprecated
	public void move(int x, int y, int z, int w) {
		super.x = x;
		super.y = y;
		this.z = z;
		this.w = w;
	}

	/**
	 * Translates this <code>Box</code> the indicated distance,
	 * to the right along the x coordinate axis, and
	 * downward along the y coordinate axis.
	 * @param x the distance to move this <code>Box</code>
	 * along the x axis
	 * @param y the distance to move this <code>Box</code>
	 * along the y axis
	 * @see java.awt.Box#setLocation(int, int)
	 * @see java.awt.Box#setLocation(java.awt.Point)
	 */
	public void translate(int x, int y, int z, int w) {
		super.x += x;
		super.y += y;
		this.z += z;
		this.w += w;
	}

	/**
	 * Gets the size of this <code>Box</code>, represented by
	 * the returned <code>Dimension</code>.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>getSize</code> method of <code>Component</code>.
	 * @return a <code>Dimension</code>, representing the size of
	 * this <code>Box</code>.
	 * @see java.awt.Component#getSize
	 * @see #setSize(Dimension)
	 * @see #setSize(int, int)
	 * @since JDK1.1
	 */
	@Override
	public Dimension4D getSize() {
		return new Dimension4D (width, height, depth, zone);
	}

	/**
	 * Sets the size of this <code>Box</code> to match the
	 * specified <code>Dimension</code>.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>setSize</code> method of <code>Component</code>.
	 * @param d the new size for the <code>Dimension</code> object
	 * @see java.awt.Component#setSize(java.awt.Dimension)
	 * @see #getSize
	 * @since JDK1.1
	 */
	public void setSize(Dimension4D d) {
		setSize(d.width, d.height, d.depth, d.zone);
	}

	/**
	 * Sets the size of this <code>Box</code> to the specified
	 * width, height, depth, and zone.
	 * <p>
	 * This method is included for completeness, to parallel the
	 * <code>setSize</code> method of <code>Component</code>.
	 * @param width the new width for this <code>Box</code>
	 * @param height the new height for this <code>Box</code>
	 * @param depth the new depth for this <code>Box</code>
	 * @param zone the new zone for this <code>Box</code>
	 * @see java.awt.Component#setSize(int, int)
	 * @see #getSize
	 * @since JDK1.1
	 */
	public void setSize(int width, int height, int depth, int zone) {
		resize(width, height, depth, zone);
	}

	/**
	 * Sets the size of this <code>Box</code> to the specified
	 * width, height, depth, and zone.
	 * <p>
	 * @param width the new width for this <code>Box</code>
	 * @param height the new height for this <code>Box</code>
	 * @param depth the new depth for this <code>Box</code>
	 * @param zone the new zone for this <code>Box</code>
	 * @deprecated As of JDK version 1.1,
	 * replaced by <code>setSize(int, int)</code>.
	 */
	@Deprecated
	public void resize(int width, int height, int depth, int zone) {
		super.width = width;
		super.height = height;
		this.depth = depth;
		this.zone = zone;
	}

	/**
	 * Checks whether or not this <code>Box</code> contains the
	 * specified <code>Point</code>.
	 * @param p the <code>Point</code> to test
	 * @return <code>true</code> if the <code>Point</code>
	 * (<i>x</i>,&nbsp;<i>y</i>) is inside this
	 * <code>Box</code>;
	 * <code>false</code> otherwise.
	 * @since JDK1.1
	 */
	public boolean contains(Point4D p) {
		return contains(p.x, p.y, p.z, p.w);
	}

	/**
	 * Checks whether or not this <code>Box</code> contains the
	 * point at the specified location
	 * (<i>x</i>,&nbsp;<i>y</i>).
	 * @param x the specified x coordinate
	 * @param y the specified y coordinate
	 * @return <code>true</code> if the point
	 * (<i>x</i>,&nbsp;<i>y</i>) is inside this
	 * <code>Box</code>;
	 * <code>false</code> otherwise.
	 * @since JDK1.1
	 */
	@Override
	public boolean contains(int x, int y, int z, int w) {
		return inside(x, y, z, w);
	}

	/**
	 * Checks whether or not this <code>Box</code> entirely contains
	 * the specified <code>Box</code>.
	 *
	 * @param r the specified <code>Box</code>
	 * @return <code>true</code> if the <code>Box</code>
	 * is contained entirely inside this <code>Box</code>;
	 * <code>false</code> otherwise
	 * @since JDK1.2
	 */
	public boolean contains(Box4D b) {
		return contains(b.x, b.y, b.z, b.w, b.width, b.height, b.depth, b.zone);
	}

	/**
	 * Checks whether this <code>Box</code> entirely contains
	 * the <code>Box</code>
	 * at the specified location (<i>X</i>,&nbsp;<i>Y</i>) with the
	 * specified dimensions (<i>W</i>,&nbsp;<i>H</i>).
	 * @param X the specified x coordinate
	 * @param Y the specified y coordinate
	 * @param Z the specified z coordinate
	 * @param W the specified w coordinate
	 * @param WL the width of the <code>Box</code>
	 * @param HL the height of the <code>Box</code>
	 * @param DL the depth of the <code>Box</code>
	 * @param ZL the zone of the <code>Box</code>
	 * @return <code>true</code> if the <code>Box</code> specified by
	 * (<i>X</i>,&nbsp;<i>Y</i>,&nbsp;<i>WL</i>,&nbsp;<i>HL</i>)
	 * is entirely enclosed inside this <code>Box</code>;
	 * <code>false</code> otherwise.
	 * @since JDK1.1
	 */
	public boolean contains(int X, int Y, int Z, int W, int WL, int HL, int DL, int ZL) {
		int wl = super.width;
		int hl = super.height;
		int dl = this.depth;
		int zl = this.zone;
		if ((wl | hl | dl | zl | WL | HL | DL | ZL) < 0) {
			// At least one of the dimensions is negative...
			return false;
		}
		// Note: if any dimension is zero, tests below must return false...
		int x = super.x;
		int y = super.y;
		int z = this.z;
		int w = this.w;
		if (X < x || Y < y || Z < z) {
			return false;
		}
		int WX = WL + X;
		int HY = HL + Y;
		int DZ = DL + Z;
		int ZW = ZL + Z;
		wl += x;
		if (WX <= X) {
			// X+WL overflowed or WL was zero, return false if...
			// either original w or W was zero or
			// x+w did not overflow or
			// the overflowed x+w is smaller than the overflowed X+W
			if (wl >= x || WX > w) return false;
		} else {
			// X+WL did not overflow and W was not zero, return false if...
			// original w was zero or
			// x+w did not overflow and x+w is smaller than X+WL
			if (wl >= x && WX > w) return false;
		}
		hl += y;
		if (HY <= Y) {
			if (hl >= y || HY > hl) return false;
		} else {
			if (hl >= y && HY > hl) return false;
		}
		dl += z;
		if (DZ <= Z) {
			if (dl >= z || DZ > dl) return false;
		} else {
			if (dl >= z && DZ > dl) return false;
		}
		zl += w;
		if (ZW <= W) {
			if (zl >= w || ZW > zl) return false;
		} else {
			if (zl >= w && ZW > zl) return false;
		}
		return true;
	}

	/**
	 * Checks whether or not this <code>Box</code> contains the
	 * point at the specified location
	 * (<i>X</i>,&nbsp;<i>Y</i>).
	 * @param X the specified x coordinate
	 * @param Y the specified y coordinate
	 * @return <code>true</code> if the point
	 * (<i>X</i>,&nbsp;<i>Y</i>) is inside this
	 * <code>Box</code>;
	 * <code>false</code> otherwise.
	 * @deprecated As of JDK version 1.1,
	 * replaced by <code>contains(int, int)</code>.
	 */
	@Deprecated
	public boolean inside(int X, int Y, int Z, int W) {
		int wl = super.width;
		int hl = super.height;
		int dl = this.depth;
		int zl = this.zone;
		if ((wl | hl | dl | zl) < 0) {
			// At least one of the dimensions is negative...
			return false;
		}
		// Note: if either dimension is zero, tests below must return false...
		int x = super.x;
		int y = super.y;
		if (X < x || Y < y || Z < z || W < w) {
			return false;
		}
		wl += x;
		hl += y;
		dl += z;
		zl += w;
		// overflow || intersect
		return ((wl < x || wl > X) &&
			(hl < y || hl > Y) &&
			(dl < z || dl > Z) &&
			(zl < w || zl > w));
	}

	/**
	 * Resizes the <code>Box</code> both horizontally and vertically.
	 * <p>
	 * This method modifies the <code>Box</code> so that it is
	 * <code>h</code> units larger on both the left and right side,
	 * and <code>v</code> units larger at both the top and bottom.
	 * <p>
	 * The new <code>Box</code> has (<code>x&nbsp;-&nbsp;h</code>,
	 * <code>y&nbsp;-&nbsp;v</code>) as its front-top-left corner, a
	 * width of
	 * <code>width</code>&nbsp;<code>+</code>&nbsp;<code>2h</code>,
	 * and a height of
	 * <code>height</code>&nbsp;<code>+</code>&nbsp;<code>2v</code>.
	 * <p>
	 * If negative values are supplied for <code>h</code> and
	 * <code>v</code>, the size of the <code>Box</code>
	 * decreases accordingly.
	 * The <code>grow</code> method does not check whether the resulting
	 * values of <code>width</code> and <code>height</code> are
	 * non-negative.
	 * @param h the horizontal expansion
	 * @param v the vertical expansion
	 */
	public void grow(int hl, int vl, int dl, int zl) {
		x -= hl;
		y -= vl;
		z -= dl;
		w -= zl;
		width += hl * 2;
		height += vl * 2;
		depth += dl * 2;
		zone += zl * 2;
	}

	/**
	 * Determines whether or not this <code>Box</code> is empty. A
	 * <code>Box</code> is empty if its width or its height is less
	 * than or equal to zero.
	 * @return <code>true</code> if this <code>Box</code> is empty;
	 * <code>false</code> otherwise.
	 */
	@Override
	public boolean isEmpty() {
		return (width <= 0) || (height <= 0) || (depth <= 0) || (zone <= 0);
	}

	/**
	 * Returns the hash code.
	 *
	 * @return hash code.
	 */
	@Override
	public int hashCode() {
		return super.hashCode();
	}

	/**
	 * Checks whether two boxes are equal.
	 * <p>
	 * The result is <code>true</code> if and only if the argument is not
	 * <code>null</code> and is a <code>Box</code> object that has the
	 * same front-top-left corner, width, height, and depth as this
	 * <code>Box</code>.
	 * @param obj the <code>Object</code> to compare with
	 * this <code>Box</code>
	 * @return <code>true</code> if the objects are equal;
	 * <code>false</code> otherwise.
	 */
	@Override
	public boolean equals(Object  obj) {
		if (obj instanceof Box4D ) {
			Box4D  r = (Box4D )obj;
			return ((x == r.x) &&
				(y == r.y) &&
				(z == r.z) &&
				(w == r.w) &&
				(width == r.width) &&
				(height == r.height) &&
				(depth == r.depth) &&
				(zone == r.zone));
		}
		return super.equals(obj);
	}

	/**
	 * Returns a <code>String</code> representing this
	 * <code>Box</code> and its values.
	 * @return a <code>String</code> representing this
	 * <code>Box</code> object's coordinate and size values.
	 */
	@Override
	public String toString() {
		return getClass().getName() + "[x=" + x + ",y=" + y + ",z=" + z + ",w=" + w + ",width=" + width + ",height=" + height + ",depth=" + depth + ",zone=" + zone + "]";
	}
}
