package com.sillycycle.bagleyd.util;

/*
 * @(#)Cell4D.java
 *
 * Copyright 2009 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Component;
import java.awt.Point;

/**
 * The <code>Cell4D</code> class handles the construction of the cells.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Cell4D extends Component {
	private static final long serialVersionUID = 42L;

	CellTs cellTs = null;
	CellHd cellHd = null;
	CellIt cellIt = null;

	int polychoron; /* cells of polychoron = 8, 16, 24 */

	public Cell4D(int polychoron, int nx, int ny/*, int nz, int nw*/,
			int screenWidth, int screenHeight/*,
			int mouseZone*/) {
		cellTs = null;
		//cellHd = null;
		cellIt = null;
		this.polychoron = polychoron;
		switch (polychoron) {
		case 8:
			cellTs = new CellTs(nx, ny/*, nz, nw*/,
				screenWidth, screenHeight);
			break;
		case 16:
			cellHd = new CellHd(nx, ny/*, nz, nw*/,
				screenWidth, screenHeight);
			break;
		case 24:
			cellIt = new CellIt(nx, ny/*, nz, nw*/,
				screenWidth, screenHeight);
			break;
		default:
		}
	}

	/*Box4D boundBox(int nx, int ny, int nz, int nw) {
		switch (polychoron) {
		case 8:
			return cellTs.boundBox(nx, ny, nz, nw);
		case 16:
			return cellHd.boundBox(nx, ny, nz, nw);
		case 24:
			return cellIt.boundBox(nx, ny, nz, nw);
		default:
			return null;
		}
	}*/

	public Point4D screenToCell(int x, int y, int z, int w) {
		switch (polychoron) {
		case 8:
			return new Point4D(cellTs.screenToCell(x, y, z), w);
		case 16:
			return new Point4D(cellHd.screenToCell(x, y, z), w);
		case 24:
			return new Point4D(cellIt.screenToCell(x, y, z), w);
		default:
			return null;
		}
	}

	public boolean inCell(Graphics3D g, Point pt, int i, int j, int k/*, int l*/) {
		if (g == null)
			return false;
		switch (polychoron) {
		case 8:
			return CellTs.inCell(g, pt, i, j, k/*, l*/);
		case 16:
			return CellHd.inCell(g, pt, i, j, k/*, l*/);
		case 24:
			return CellIt.inCell(g, pt, i, j, k/*, l*/);
		default:
			return false;
		}
	}

	public void drawCell(Graphics3D g, boolean round, boolean wireframe,
			Color cid, Color bid,
			int i, int j, int k/*, int l*/) {
		if (g == null)
			return;
		switch (polychoron) {
		case 8:
			CellTs.drawCell(g, round, wireframe, cid, bid, i, j, k/*, l*/);
			break;
		case 16:
			CellHd.drawCell(g, round, wireframe, cid, bid, i, j, k/*, l*/);
			break;
		case 24:
			CellIt.drawCell(g, round, wireframe, cid, bid, i, j, k/*, l*/);
			break;
		default:
		}
	}

	public void drawCaptureBox(Graphics3D g, Color color,
			int i, int j, int k, int l, int nx, int ny, int nz, int nw) {
		if (g == null)
			return;
		switch (polychoron) {
		case 8:
			CellTs.drawCaptureBox(g, color, i, j, k, l, nx, ny, nz, nw);
			break;
		case 16:
			CellHd.drawCaptureBox(g, color, i, j, k, l, nx, ny, nz, nw);
			break;
		case 24:
			CellIt.drawCaptureBox(g, color, i, j, k, l, nx, ny, nz, nw);
			break;
		default:
		}
	}
}
